import inspect
import math
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional

import mlx.core as mx
import mlx.nn as nn
from mlx_lm.models.base import (
    create_attention_mask,
    create_ssm_mask,
    scaled_dot_product_attention,
)
from PIL import Image


@dataclass
class LanguageModelOutput:
    logits: mx.array
    hidden_states: Optional[List[mx.array]] = None
    cross_attention_states: Optional[List[mx.array]] = None
    encoder_outputs: Optional[List[mx.array]] = None


@dataclass
class InputEmbeddingsFeatures:
    inputs_embeds: mx.array
    attention_mask_4d: Optional[mx.array] = None
    visual_pos_masks: Optional[mx.array] = None
    deepstack_visual_embeds: Optional[mx.array] = None
    per_layer_inputs: Optional[mx.array] = None
    cross_attention_states: Optional[mx.array] = None
    cross_attention_mask: Optional[mx.array] = None
    full_text_row_masked_out_mask: Optional[mx.array] = None
    decoder_inputs_embeds: Optional[mx.array] = None
    attention_mask: Optional[mx.array] = None  # For encoder-decoder models

    def to_dict(self):
        return {
            "inputs_embeds": self.inputs_embeds,
            "attention_mask_4d": self.attention_mask_4d,
            "visual_pos_masks": self.visual_pos_masks,
            "deepstack_visual_embeds": self.deepstack_visual_embeds,
            "per_layer_inputs": self.per_layer_inputs,
            "cross_attention_states": self.cross_attention_states,
            "cross_attention_mask": self.cross_attention_mask,
            "full_text_row_masked_out_mask": self.full_text_row_masked_out_mask,
            "decoder_inputs_embeds": self.decoder_inputs_embeds,
            "attention_mask": self.attention_mask,
        }


@dataclass
class BaseModelConfig:
    @classmethod
    def from_dict(cls, params):
        return cls(
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            }
        )

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}


class BaseImageProcessor:
    """
    Base image processor class. Subclasses should implement preprocess().
    Transformers imports are deferred to __init__ for faster module loading.
    """

    def __init__(
        self,
        image_mean=(0.5, 0.5, 0.5),
        image_std=(0.5, 0.5, 0.5),
        size=(384, 384),
        crop_size: Dict[str, int] = None,
        resample=None,
        rescale_factor=1 / 255,
        data_format=None,
    ):
        from transformers.image_processing_utils import get_size_dict
        from transformers.image_utils import ChannelDimension, PILImageResampling

        if resample is None:
            resample = PILImageResampling.BICUBIC
        if data_format is None:
            data_format = ChannelDimension.FIRST

        crop_size = (
            crop_size if crop_size is not None else {"height": 384, "width": 384}
        )
        crop_size = get_size_dict(
            crop_size, default_to_square=True, param_name="crop_size"
        )

        self.image_mean = image_mean
        self.image_std = image_std
        self.size = size
        self.resample = resample
        self.rescale_factor = rescale_factor
        self.data_format = data_format
        self.crop_size = crop_size

    def rescale(
        self,
        image,
        scale: float,
        input_data_format: str = "channels_first",
    ):
        """Rescale an image by a scale factor."""
        return image * scale

    def normalize(
        self,
        image,
        mean,
        std,
        input_data_format: str = "channels_first",
    ):
        """Normalize an image with mean and std."""
        import numpy as np

        mean = np.array(mean, dtype=image.dtype)
        std = np.array(std, dtype=image.dtype)

        if input_data_format == "channels_first":
            # Image shape: [C, H, W]
            mean = mean[:, None, None]
            std = std[:, None, None]
        else:
            # Image shape: [H, W, C]
            pass  # mean and std are already in correct shape

        return (image - mean) / std

    @abstractmethod
    def preprocess(self, images):
        pass


def expand2square(pil_img, background_color):
    width, height = pil_img.size
    if width == height:
        return pil_img
    elif width > height:
        result = Image.new(pil_img.mode, (width, width), background_color)
        result.paste(pil_img, (0, (width - height) // 2))
        return result
    else:
        result = Image.new(pil_img.mode, (height, height), background_color)
        result.paste(pil_img, ((height - width) // 2, 0))
        return result


def check_array_shape(arr):
    shape = arr.shape

    # Check if the shape has 4 dimensions
    if len(shape) == 4:
        out_channels, kH, KW, _ = shape
        # Check if out_channels is the largest, and kH and KW are the same
        if (out_channels >= kH) and (out_channels >= KW) and (kH == KW):
            return True
        else:
            return False
    # Check if the shape has 3 dimensions
    elif len(shape) == 3:
        _, kW, out_channels = shape
        # Check if out_channels is the largest
        if kW >= out_channels:
            return True
        else:
            return False
    else:
        return False


def check_activation_stats(name, tensor):
    """Helper function to check for anomalies and log stats."""

    print(f"--- Activation Stats: {name} ---")
    # Check for NaNs/Infs
    has_nan = mx.isnan(tensor).any()
    has_inf = mx.isinf(tensor).any()
    if has_nan:
        print(f"WARNING: Found NaN in {name}")
    if has_inf:
        print(f"WARNING: Found Inf in {name}")

    # Calculate and print stats (ensure computation happens)
    min_val = mx.min(tensor).item()
    max_val = mx.max(tensor).item()
    mean_val = mx.mean(tensor).item()
    std_val = mx.std(tensor).item()
    print(f"  Shape: {tensor.shape}")
    print(f"  Min: {min_val:.4f}, Max: {max_val:.4f}")
    print(f"  Mean: {mean_val:.4f}, Std: {std_val:.4f}")
    print("-" * (len(name) + 24))


def pixel_shuffle(input_tensor, shuffle_ratio):
    # input_tensor: [batch_size, num_patches, channels]
    batch_size, num_patches, channels = input_tensor.shape
    patch_size = int(math.sqrt(num_patches))

    input_tensor = input_tensor.reshape(batch_size, patch_size, patch_size, -1)
    batch_size, height, width, channels = input_tensor.shape

    reshaped_tensor = input_tensor.reshape(
        batch_size, height, int(width * shuffle_ratio), int(channels / shuffle_ratio)
    )
    reshaped_tensor = reshaped_tensor.transpose(0, 2, 1, 3)

    reshaped_tensor = reshaped_tensor.reshape(
        batch_size,
        int(height * shuffle_ratio),
        int(width * shuffle_ratio),
        int(channels / (shuffle_ratio**2)),
    )
    reshaped_tensor = reshaped_tensor.transpose(0, 2, 1, 3)

    output_tensor = reshaped_tensor.reshape(batch_size, -1, reshaped_tensor.shape[-1])
    return output_tensor


def interpolate(pos_embed, size, mode="cubic", align_corners=False):
    """
    MLX implementation of PyTorch's F.interpolate with bicubic mode

    Args:
        pos_embed: MLX array with shape [B, C, H_src, W_src] or [C, H_src, W_src]
        size: Tuple (H_dst, W_dst) - target size
        align_corners: Boolean - whether to align corners

    Returns:
        Interpolated array with shape [B, C, H_dst, W_dst] or [C, H_dst, W_dst]
    """
    # Handle different input shapes
    input_dim = pos_embed.ndim
    original_shape = pos_embed.shape

    if input_dim == 3:
        # [C, H, W] -> [1, C, H, W]
        pos_embed = pos_embed.reshape(1, *original_shape)

    # Get source dimensions
    h_src, w_src = pos_embed.shape[-2:]
    h_dst, w_dst = size

    # Calculate scale factors
    scale_h = h_dst / h_src
    scale_w = w_dst / w_src

    # Create upsampler
    upsampler = nn.Upsample(
        scale_factor=(scale_h, scale_w), mode=mode, align_corners=align_corners
    )

    # Apply upsampling
    result = upsampler(pos_embed)

    # Return in the original dimension format
    if input_dim == 3:
        return result.reshape(original_shape[0], *size)
    return result


@mx.compile
def chunked_attention(
    queries: mx.array,
    keys: mx.array,
    values: mx.array,
    scale: float,
    chunk_size: int,
) -> mx.array:

    L = queries.shape[2]

    outputs = []
    for i in range(0, L, chunk_size):
        end_idx = min(i + chunk_size, L)
        q_chunk = queries[:, :, i:end_idx, :]  # (B, n_heads, chunk, head_dim)

        chunk_output = mx.fast.scaled_dot_product_attention(
            q_chunk, keys, values, scale=scale
        )

        outputs.append(chunk_output)

    return mx.concatenate(outputs, axis=2)  # (B, n_heads, L, head_dim)


def install_auto_processor_patch(target_model_types, processor_cls):
    """
    Install a composable patch on transformers.AutoProcessor.from_pretrained

    Args:
        target_model_types (Union[str, List[str]]): Model types to intercept.
        processor_cls (type): Processor class exposing `from_pretrained`.

    Returns:
        The previous `AutoProcessor.from_pretrained` for reference.
    """
    from transformers import AutoProcessor as _HF_AutoProcessor

    if isinstance(target_model_types, str):
        target_model_types = [target_model_types]
    target_model_types = {t.lower() for t in target_model_types}

    previous_from_pretrained = _HF_AutoProcessor.from_pretrained

    @classmethod
    def _patched_auto_processor_from_pretrained(
        cls, pretrained_model_name_or_path, **kwargs
    ):
        import json as _json
        from pathlib import Path

        try:
            model_path = Path(pretrained_model_name_or_path)
            is_local = model_path.exists() and model_path.is_dir()

            cfg = {}
            if is_local:
                config_path = model_path / "config.json"
                if config_path.exists():
                    with open(config_path, "r", encoding="utf-8") as f:
                        cfg = _json.load(f)
            else:
                try:
                    from huggingface_hub import hf_hub_download

                    cfg_path = hf_hub_download(
                        pretrained_model_name_or_path, "config.json"
                    )
                    with open(cfg_path, "r", encoding="utf-8") as f:
                        cfg = _json.load(f)
                except Exception:
                    cfg = {}

            model_type = str(cfg.get("model_type", "")).lower()
            if model_type in target_model_types:
                return processor_cls.from_pretrained(
                    pretrained_model_name_or_path, **kwargs
                )
        except Exception:
            # On any failure, fall back to previous behavior
            pass

        # Chain to the prior from_pretrained (which may already be patched)
        return previous_from_pretrained.__func__(
            cls, pretrained_model_name_or_path, **kwargs
        )

    _HF_AutoProcessor.from_pretrained = _patched_auto_processor_from_pretrained
    return previous_from_pretrained



import argparse
import csv
import json
import logging
import os
import random
import re
from pathlib import Path
from typing import Optional

from datasets import load_dataset
from PIL import Image
from tqdm import tqdm

from mlx_vlm import load
from mlx_vlm.evals.utils import inference


def process_question(sample: dict) -> str:
    """Format the question with choices if it's multiple choice."""
    question = sample["query"]

    if sample["question_type"] == "multi_choice" and sample["choices"]:
        choices_text = "\n".join(
            [f"({chr(65+i)}) {choice}" for i, choice in enumerate(sample["choices"])]
        )
        question = f"{question}\n{choices_text}"
    return question


def normalize_answer(response: str, problem: dict) -> Optional[str]:
    """Normalize the model's response to extract the answer."""
    response = response.strip()

    if not response:
        return None

    question_type = problem["question_type"]
    answer_type = problem["answer_type"]
    choices = problem.get("choices", [])

    # For multiple choice, try to extract the letter
    if question_type == "multi_choice":
        # First, try to find boxed answers
        boxed_match = re.search(r"\\boxed\{([^}]+)\}", response)
        if boxed_match:
            boxed_content = boxed_match.group(1)
            # Check if it's a choice letter
            letter_match = re.match(
                r"^\(?([A-Z])\)?\.?$", boxed_content.strip().upper()
            )
            if letter_match:
                letter = letter_match.group(1)
                idx = ord(letter) - ord("A")
                if 0 <= idx < len(choices):
                    return choices[idx]
            # Check if it's directly one of the choices
            if boxed_content.strip() in choices:
                return boxed_content.strip()

        # Try to find Chinese answer pattern "故选：X" or "故选X"
        chinese_match = re.search(r"故选[：:]\s*([A-Z])", response.upper())
        if not chinese_match:
            chinese_match = re.search(r"故选\s*([A-Z])", response.upper())
        if chinese_match:
            letter = chinese_match.group(1)
            idx = ord(letter) - ord("A")
            if 0 <= idx < len(choices):
                return choices[idx]

        # Try to find "the answer is X" or "answer: X" patterns near the end
        answer_patterns = [
            r"(?:the\s+)?answer\s+is\s+\(?([A-Z])\)?",
            r"answer:\s*\(?([A-Z])\)?",
            r"choose\s+\(?([A-Z])\)?",
            r"option\s+\(?([A-Z])\)?",
        ]

        # Search from the end of the response (last 500 chars)
        end_section = response[-500:] if len(response) > 500 else response
        for pattern in answer_patterns:
            matches = list(re.finditer(pattern, end_section, re.IGNORECASE))
            if matches:
                # Take the last match
                letter = matches[-1].group(1).upper()
                idx = ord(letter) - ord("A")
                if 0 <= idx < len(choices):
                    return choices[idx]

        # Look for patterns like "(A)", "A)", "A.", "A" - prioritize from the end
        matches = list(re.finditer(r"\(?([A-Z])\)?\.?", response.upper()))
        if matches:
            # Try the last few matches first
            for match in reversed(matches[-5:]):
                letter = match.group(1)
                idx = ord(letter) - ord("A")
                if 0 <= idx < len(choices):
                    return choices[idx]

        # If the response is exactly one of the choices
        if response in choices:
            return response

        # Try to find the most similar choice using edit distance
        def distance(s1, s2):
            if len(s1) < len(s2):
                return distance(s2, s1)
            if len(s2) == 0:
                return len(s1)

            previous_row = range(len(s2) + 1)
            for i, c1 in enumerate(s1):
                current_row = [i + 1]
                for j, c2 in enumerate(s2):
                    insertions = previous_row[j + 1] + 1
                    deletions = current_row[j] + 1
                    substitutions = previous_row[j] + (c1 != c2)
                    current_row.append(min(insertions, deletions, substitutions))
                previous_row = current_row

            return previous_row[-1]

        if choices:
            distances = [
                distance(response.lower(), choice.lower()) for choice in choices
            ]
            return choices[distances.index(min(distances))]

    # For integer answers
    elif answer_type == "integer":
        # First try to find boxed answer
        boxed_match = re.search(r"\\boxed\{([^}]+)\}", response)
        if boxed_match:
            boxed_content = boxed_match.group(1)
            # Remove commas from numbers
            boxed_content = boxed_content.replace(",", "")
            # Try scientific notation first
            sci_numbers = re.findall(r"-?\d+\.?\d*[eE][+-]?\d+", boxed_content)
            if sci_numbers:
                try:
                    return str(int(float(sci_numbers[0])))
                except:
                    pass
            # Then regular numbers
            numbers = re.findall(r"-?\d+", boxed_content)
            if numbers:
                try:
                    return str(int(numbers[0]))
                except:
                    pass

        # Try common answer patterns near the end
        end_section = response[-500:] if len(response) > 500 else response
        answer_patterns = [
            r"(?:the\s+)?answer\s+is\s+(-?[\d,]+\.?\d*[eE][+-]?\d+|-?[\d,]+)",
            r"answer:\s*(-?[\d,]+\.?\d*[eE][+-]?\d+|-?[\d,]+)",
            r"(?:total|result|left|remaining)(?:\s+is|\s+are|:)\s*(-?[\d,]+\.?\d*[eE][+-]?\d+|-?[\d,]+)",
        ]

        for pattern in answer_patterns:
            matches = list(re.finditer(pattern, end_section, re.IGNORECASE))
            if matches:
                try:
                    # Remove commas before converting
                    num_str = matches[-1].group(1).replace(",", "")
                    return str(int(float(num_str)))
                except:
                    pass

        # Look for scientific notation anywhere in response
        sci_numbers = re.findall(r"-?\d+\.?\d*[eE][+-]?\d+", response)
        if sci_numbers:
            try:
                return str(int(float(sci_numbers[-1])))
            except:
                pass

        # Fall back to finding all numbers (including comma-formatted) and taking the last one
        # Match numbers with optional commas: 7,518 or 7518
        numbers = re.findall(r"-?[\d,]+", response)
        if numbers:
            try:
                # Remove commas and try the last number first
                return str(int(numbers[-1].replace(",", "")))
            except:
                pass

    # For float answers
    elif answer_type == "float":
        precision = int(problem.get("precision", 2))

        # First try to find boxed answer
        boxed_match = re.search(r"\\boxed\{([^}]+)\}", response)
        if boxed_match:
            boxed_content = boxed_match.group(1)
            # Try scientific notation first
            sci_numbers = re.findall(r"-?\d+\.?\d*[eE][+-]?\d+", boxed_content)
            if sci_numbers:
                try:
                    return str(round(float(sci_numbers[0]), precision))
                except:
                    pass
            # Then regular numbers
            numbers = re.findall(r"-?\d+\.?\d*", boxed_content)
            if numbers:
                try:
                    return str(round(float(numbers[0]), precision))
                except:
                    pass

        # Try common answer patterns near the end
        end_section = response[-500:] if len(response) > 500 else response
        answer_patterns = [
            r"(?:the\s+)?answer\s+is\s+(-?\d+\.?\d*[eE][+-]?\d+|-?\d+\.?\d*)",
            r"answer:\s*(-?\d+\.?\d*[eE][+-]?\d+|-?\d+\.?\d*)",
            r"d\s*=\s*(-?\d+\.?\d*[eE][+-]?\d+|-?\d+\.?\d*)",  # For physics problems with d=
        ]

        for pattern in answer_patterns:
            matches = list(re.finditer(pattern, end_section, re.IGNORECASE))
            if matches:
                try:
                    return str(round(float(matches[-1].group(1)), precision))
                except:
                    pass

        # Look for scientific notation anywhere in response
        sci_numbers = re.findall(r"-?\d+\.?\d*[eE][+-]?\d+", response)
        if sci_numbers:
            try:
                return str(round(float(sci_numbers[-1]), precision))
            except:
                pass

        # Fall back to finding all numbers and taking the last one
        numbers = re.findall(r"-?\d+\.?\d*", response)
        if numbers:
            try:
                # Try the last number first
                return str(round(float(numbers[-1]), precision))
            except:
                pass

    return response


def evaluate_answer(prediction: Optional[str], ground_truth: str) -> bool:
    """Check if the prediction matches the ground truth."""
    if prediction is None:
        return False
    try:
        # First check exact match
        if str(prediction).strip() == str(ground_truth).strip():
            return True

        # Handle numeric word representations
        word_to_num = {
            "zero": "0",
            "one": "1",
            "two": "2",
            "three": "3",
            "four": "4",
            "five": "5",
            "six": "6",
            "seven": "7",
            "eight": "8",
            "nine": "9",
            "ten": "10",
            "eleven": "11",
            "twelve": "12",
            "thirteen": "13",
            "fourteen": "14",
            "fifteen": "15",
            "sixteen": "16",
            "seventeen": "17",
            "eighteen": "18",
            "nineteen": "19",
            "twenty": "20",
        }

        pred_normalized = str(prediction).strip().lower()
        gt_normalized = str(ground_truth).strip().lower()

        # Convert words to numbers
        if pred_normalized in word_to_num:
            pred_normalized = word_to_num[pred_normalized]
        if gt_normalized in word_to_num:
            gt_normalized = word_to_num[gt_normalized]

        return pred_normalized == gt_normalized
    except:
        return False


def parse_args():
    parser = argparse.ArgumentParser(
        description="Evaluate models on MathVista benchmark"
    )
    parser.add_argument(
        "--model",
        type=str,
        required=True,
        help="The path to the MLX VLM model",
    )
    parser.add_argument(
        "--adapter-path",
        type=str,
        help="Optional path for the trained adapter weights and config",
    )
    parser.add_argument(
        "--dataset",
        type=str,
        default="AI4Math/MathVista",
        help="Hugging Face dataset name",
    )
    parser.add_argument(
        "--split",
        type=str,
        default="testmini",
        choices=["testmini", "test"],
        help="Dataset split to evaluate on",
    )
    parser.add_argument(
        "--streaming",
        action="store_true",
        help="Use streaming dataset loading",
    )
    parser.add_argument(
        "--max-samples",
        type=int,
        default=None,
        help="Maximum number of samples to evaluate (for debugging)",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        default="results/mathvista",
        help="Directory to save results",
    )
    parser.add_argument(
        "--max-tokens",
        type=int,
        default=512,
        help="Maximum number of tokens to generate",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=0.0,
        help="Temperature for generation",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Print detailed output for debugging",
    )
    parser.add_argument("--seed", type=int, default=42, help="Random seed")
    return parser.parse_args()


def main():
    args = parse_args()

    random.seed(args.seed)

    # Setup logging
    logging.basicConfig(
        level=logging.INFO if args.verbose else logging.WARNING,
        format="%(asctime)s - %(levelname)s - %(message)s",
    )

    logging.info(f"Loading model from {args.model}")
    model, processor = load(
        args.model, adapter_path=args.adapter_path, trust_remote_code=True
    )

    # Load dataset
    logging.info(f"Loading dataset {args.dataset}, split {args.split}")
    dataset = load_dataset(args.dataset, split=args.split, streaming=args.streaming)

    if args.max_samples:
        dataset = dataset.select(range(min(args.max_samples, len(dataset))))

    # Create output directory
    output_dir = Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    results = {}
    category_scores = {}
    correct = 0
    total = 0

    # Evaluate each sample
    for idx, sample in enumerate(tqdm(dataset, desc="Evaluating")):
        pid = sample["pid"]

        try:
            # Load and process image
            if "decoded_image" in sample and sample["decoded_image"]:
                if isinstance(sample["decoded_image"], str):
                    image_path = sample["decoded_image"]
                    if os.path.exists(image_path):
                        image = Image.open(image_path).convert("RGB")
                    else:
                        logging.warning(
                            f"Image not found: {image_path}, skipping sample {pid}"
                        )
                        continue
                else:
                    # Image is already loaded
                    image = sample["decoded_image"].convert("RGB")
            else:
                logging.warning(f"No image for sample {pid}, skipping")
                continue

            # Create prompt
            prompt = process_question(sample)

            # Generate response
            output = inference(
                model,
                processor,
                prompt,
                image=image,
                max_tokens=args.max_tokens,
                temperature=args.temperature,
            )

            response = output.strip()

            # Normalize answer
            prediction = normalize_answer(response, sample)

            # Evaluate
            ground_truth = sample.get("answer", "")
            if args.split == "testmini" and ground_truth:
                is_correct = evaluate_answer(prediction, ground_truth)
                if is_correct:
                    correct += 1
            else:
                is_correct = None

            total += 1

            # Store results
            results[pid] = {
                "pid": pid,
                "question": sample["question"],
                "query": sample["query"],
                "question_type": sample["question_type"],
                "answer_type": sample["answer_type"],
                "choices": sample.get("choices", []),
                "unit": sample.get("unit", ""),
                "precision": sample.get("precision", 0),
                "ground_truth": ground_truth,
                "response": response,
                "prediction": prediction,
                "correct": is_correct,
                "metadata": sample.get("metadata", {}),
            }
            # Track category-wise performance
            category = sample.get("metadata", {}).get("category", "unknown")
            if category not in category_scores:
                category_scores[category] = {"correct": 0, "total": 0}

            category_scores[category]["total"] += 1
            if is_correct:
                category_scores[category]["correct"] += 1

            if args.verbose:
                logging.info(f"\nSample {pid}:")
                logging.info(f"Question: {sample['question']}")
                logging.info(f"Response: {response}")
                logging.info(f"Prediction: {prediction}")
                logging.info(f"Ground Truth: {ground_truth}")
                logging.info(f"Correct: {is_correct}")

        except Exception as e:
            logging.error(f"Error processing sample {pid}: {e}")
            continue

    # Calculate accuracy if applicable
    if args.split == "testmini":
        accuracy = correct / total if total > 0 else 0
    else:
        accuracy = None
        correct = None

    # Save results
    model_name = args.model.split("/")[-1]
    results_file = output_dir / f"{model_name}_MathVista_{args.split}.csv"

    # Convert results to list of dictionaries for CSV writing
    fieldnames = [
        "pid",
        "question",
        "query",
        "question_type",
        "answer_type",
        "choices",
        "unit",
        "precision",
        "ground_truth",
        "response",
        "prediction",
        "correct",
        "metadata",
    ]

    with open(results_file, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writeheader()

        for result in results.values():
            # Convert list and dict fields to strings for CSV
            row = result.copy()
            if isinstance(row.get("choices"), list):
                row["choices"] = "; ".join(row["choices"])
            if isinstance(row.get("metadata"), dict):
                row["metadata"] = json.dumps(row["metadata"])
            writer.writerow(row)

    # Save summary
    summary = {
        "model": args.model,
        "dataset": args.dataset,
        "split": args.split,
        "total_samples": total,
        "category_scores": category_scores,
    }

    if accuracy is not None:
        summary["correct"] = correct
        summary["accuracy"] = accuracy

    summary_file = output_dir / f"{model_name}_MathVista_{args.split}.json"
    with open(summary_file, "w") as f:
        json.dump(summary, f, indent=2)

    print(f"\n{'='*80}")
    print("MathVista Evaluation Results")
    print(f"{'='*80}")
    print(f"Model: {args.model}")
    print(f"Split: {args.split}")
    print(f"Total Samples: {total}")
    if accuracy is not None:
        print(f"Correct: {correct}")
        print(f"Accuracy: {accuracy*100:.2f}%")
    else:
        print("Accuracy not computed for this split (no ground truth labels)")

    print("\n" + "-" * 80)
    print(f"Subcategory Scores:")
    print(f"{'-'*80}")
    for category, scores in category_scores.items():
        cat_total = scores["total"]
        cat_correct = scores["correct"]
        cat_accuracy = cat_correct / cat_total if cat_total > 0 else 0
        print(f"  {category}: {cat_correct}/{cat_total} ({cat_accuracy*100:.2f}%)")
    print(f"{'='*80}")
    print(f"\nResults saved to {results_file} and {summary_file}")


if __name__ == "__main__":
    main()



from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from ..base import InputEmbeddingsFeatures
from .config import ModelConfig
from .language import LanguageModel
from .processing import Glm46VProcessor
from .vision import VisionModel

# Register the processor with the name expected by the model config
try:
    from transformers import AutoProcessor

    # The model's preprocessor_config.json specifies "processor_class": "Glm46VProcessor"
    AutoProcessor.register("Glm46VProcessor", Glm46VProcessor)
except Exception as e:
    print(f"Error registering glm4v processor: {e}")


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config, config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        image_grid_thw = kwargs.pop("image_grid_thw", None)
        video_grid_thw = kwargs.pop("video_grid_thw", None)
        mask = kwargs.get("mask", None)
        grid_thw = image_grid_thw if image_grid_thw is not None else video_grid_thw

        if pixel_values is None:
            # Reset position state for text-only generation
            self.language_model._position_ids = None
            self.language_model._rope_deltas = None
            return InputEmbeddingsFeatures(
                inputs_embeds=self.language_model.model.embed_tokens(input_ids)
            )

        dtype = self.vision_tower.patch_embed.proj.weight.dtype
        pixel_values = pixel_values.astype(dtype)

        # Get the input embeddings from the language model
        inputs_embeds = self.language_model.model.embed_tokens(input_ids)

        # Get the ouptut hidden states from the vision model
        hidden_states = self.vision_tower(
            pixel_values, grid_thw, output_hidden_states=False
        )

        split_sizes = (
            image_grid_thw.prod(-1) // self.vision_tower.spatial_merge_size**2
        ).tolist()
        hidden_states = mx.split(
            hidden_states, [split_sizes[0], sum(split_sizes[:2])], axis=0
        )

        hidden_states = mx.concatenate(hidden_states, axis=0).astype(
            hidden_states[0].dtype
        )

        # Insert special image tokens in the input_ids
        final_inputs_embeds = self.merge_input_ids_with_image_features(
            self.config.image_token_id,
            self.config.video_token_id,
            hidden_states,
            inputs_embeds,
            input_ids,
        )

        # Pre-calculate position_ids for chunked prefill
        if image_grid_thw is not None or video_grid_thw is not None:
            position_ids, rope_deltas = self.language_model.get_rope_index(
                input_ids, image_grid_thw, video_grid_thw, mask
            )
            self.language_model._position_ids = position_ids
            self.language_model._rope_deltas = rope_deltas

        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    @staticmethod
    def merge_input_ids_with_image_features(
        image_token_id,
        video_token_id,
        image_features,
        inputs_embeds,
        input_ids,
    ):
        """Merge image features into input embeddings at image token positions.

        Args:
            image_token_id: The token ID for image placeholders
            video_token_id: The token ID for video placeholders (fallback)
            image_features: Vision features from the vision tower [num_features, hidden_dim]
            inputs_embeds: Input embeddings [batch_size, seq_len, hidden_dim]
            input_ids: Input token IDs [batch_size, seq_len]
            grid_thw: Grid dimensions for each image (optional, not used in simple case)

        Returns:
            Updated input embeddings with image features inserted
        """
        # Find positions of image tokens
        image_positions = input_ids == image_token_id
        if mx.sum(image_positions) == 0:
            image_positions = input_ids == video_token_id

        # Get dimensions
        batch_size, seq_len = input_ids.shape

        # Process each batch item
        batch_outputs = []
        feature_start_idx = 0

        for batch_idx in range(batch_size):
            # Get mask for this batch
            image_mask = image_positions[batch_idx]
            num_positions = mx.sum(image_mask).item()

            if num_positions > 0:
                # Extract features for this batch
                batch_features = image_features[
                    feature_start_idx : feature_start_idx + num_positions
                ]

                # Validate we have the right number of features
                if batch_features.shape[0] != num_positions:
                    raise ValueError(
                        f"Number of image token positions ({num_positions}) does not match "
                        f"number of image features ({batch_features.shape[0]}) for batch {batch_idx}"
                    )

                # Create indices for gathering
                cumsum = mx.cumsum(image_mask.astype(mx.int32))
                feature_indices = mx.where(image_mask, cumsum - 1, 0)

                # Gather features
                gathered_features = batch_features[feature_indices]

                # Combine with original embeddings
                image_mask_expanded = mx.expand_dims(image_mask, axis=-1)
                batch_output = mx.where(
                    image_mask_expanded, gathered_features, inputs_embeds[batch_idx]
                )

                feature_start_idx += num_positions
            else:
                # No image tokens in this batch item
                batch_output = inputs_embeds[batch_idx]

            batch_outputs.append(batch_output)

        # Stack all batch outputs
        return mx.stack(batch_outputs, axis=0)

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):

        input_embeddings_features = self.get_input_embeddings(
            input_ids, pixel_values, **kwargs
        )

        logits = self.language_model(
            input_ids,
            input_embeddings_features.inputs_embeds,
            mask=mask,
            cache=cache,
            **kwargs,
        )

        return logits

    def sanitize(self, weights):
        def transform_key(key):
            if "visual" in key:
                if "vision_tower" not in key:
                    key = key.replace("model.", "").replace("visual", "vision_tower")
            if "model.language_model" in key:
                key = key.replace("model.language_model", "language_model.model")
            if "lm_head" in key and not key.startswith("language_model"):
                key = key.replace("lm_head", "language_model.lm_head")
            return key

        return {transform_key(k): v for k, v in weights.items()}



from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from ..base import InputEmbeddingsFeatures
from ..llava_bunny.language import LanguageModel
from .config import ModelConfig
from .vision import VisionModel


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        if pixel_values is None:
            return InputEmbeddingsFeatures(
                inputs_embeds=self.language_model.model.embed_tokens(input_ids)
            )

        image_grid_thw = kwargs.get("image_grid_thw", None)
        if image_grid_thw is None:
            raise ValueError("image_grid_thw is required when pixel_values is provided")

        dtype = self.vision_tower.patch_embed.patchifier.proj.weight.dtype
        pixel_values = pixel_values.astype(dtype)

        inputs_embeds = self.language_model.model.embed_tokens(input_ids)
        hidden_states = self.vision_tower(
            pixel_values, image_grid_thw, output_hidden_states=False
        )

        final_inputs_embeds = self.merge_input_ids_with_image_features(
            self.config.image_token_id,
            self.config.video_token_id,
            hidden_states,
            inputs_embeds,
            input_ids,
        )

        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    @staticmethod
    def merge_input_ids_with_image_features(
        image_token_id,
        video_token_id,
        image_features,
        inputs_embeds,
        input_ids,
    ):
        image_positions = input_ids == image_token_id
        if mx.sum(image_positions) == 0:
            image_positions = input_ids == video_token_id

        batch_outputs = []
        feature_start_idx = 0

        for batch_idx in range(input_ids.shape[0]):
            image_mask = image_positions[batch_idx]
            num_positions = mx.sum(image_mask).item()

            if num_positions > 0:
                batch_features = image_features[
                    feature_start_idx : feature_start_idx + num_positions
                ]

                if batch_features.shape[0] != num_positions:
                    raise ValueError(
                        f"Number of image token positions ({num_positions}) does not match "
                        f"number of image features ({batch_features.shape[0]}) for batch {batch_idx}"
                    )

                cumsum = mx.cumsum(image_mask.astype(mx.int32))
                feature_indices = mx.where(image_mask, cumsum - 1, 0)
                gathered_features = batch_features[feature_indices]

                image_mask_expanded = mx.expand_dims(image_mask, axis=-1)
                batch_output = mx.where(
                    image_mask_expanded, gathered_features, inputs_embeds[batch_idx]
                )

                feature_start_idx += num_positions
            else:
                batch_output = inputs_embeds[batch_idx]

            batch_outputs.append(batch_output)

        return mx.stack(batch_outputs, axis=0)

    @property
    def layers(self):
        return self.language_model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):
        input_embeddings_features = self.get_input_embeddings(
            input_ids=input_ids,
            pixel_values=pixel_values,
            **kwargs,
        )

        logits = self.language_model(
            input_ids,
            inputs_embeds=input_embeddings_features.inputs_embeds,
            mask=mask,
            cache=cache,
        )
        return logits

    def sanitize(self, weights):
        def transform_key(key):
            if key.startswith("model.vision_tower."):
                key = key.replace("model.vision_tower.", "vision_tower.", 1)
            elif key.startswith("model."):
                key = key.replace("model.", "language_model.model.", 1)
            elif key.startswith("lm_head."):
                key = key.replace("lm_head.", "language_model.model.lm_head.", 1)
            return key

        return {transform_key(k): v for k, v in weights.items()}



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from mlx_vlm.models.qwen3_omni_moe.code2wav import Code2WavModel
from mlx_vlm.models.qwen3_omni_moe.talker import Talker
from mlx_vlm.models.qwen3_omni_moe.thinker import Thinker

from .config import ModelConfig


def masked_scatter(
    final_embedding: mx.array,
    image_mask_expanded: mx.array,
    scaled_image_features: mx.array,
):
    final_embedding_shape = final_embedding.shape
    scaled_image_features_flattened = mx.flatten(scaled_image_features)
    final_embedding_flattened = mx.flatten(final_embedding)
    image_mask_expanded_flattened = mx.flatten(image_mask_expanded)

    image_positions = mx.array(np.where(image_mask_expanded_flattened)[0], mx.uint32)
    final_embedding_flattened[image_positions] = scaled_image_features_flattened

    final_embedding = mx.reshape(final_embedding_flattened, final_embedding_shape)

    return final_embedding


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config

        self.thinker = Thinker(config.thinker_config)
        self.has_talker = config.enable_audio_output
        if self.has_talker:
            self.talker = Talker(config.talker_config)
            self.code2wav = Code2WavModel(config.code2wav_config)
        else:
            self.talker = None
            self.code2wav = None

    def enable_talker(self):
        if not self.has_talker:
            self.talker = Talker(self.config.talker_config)
            self.code2wav = Code2WavModel(self.config.code2wav_config)
            self.has_talker = True

    def disable_talker(self):
        if self.has_talker:
            self.talker = None
            self.code2wav = None
            self.has_talker = False

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        pixel_values_videos: Optional[mx.array] = None,
        input_features: Optional[mx.array] = None,
        input_features_mask: Optional[mx.array] = None,
        image_grid_thw: Optional[mx.array] = None,
        video_grid_thw: Optional[mx.array] = None,
        audio_feature_lengths: Optional[mx.array] = None,
        **kwargs,
    ):
        return self.thinker.get_input_embeddings(
            input_ids=input_ids,
            pixel_values=pixel_values,
            pixel_values_videos=pixel_values_videos,
            image_grid_thw=image_grid_thw,
            video_grid_thw=video_grid_thw,
            input_features=input_features,
            feature_attention_mask=input_features_mask,
            audio_feature_lengths=audio_feature_lengths,
        )

    def get_audio_features(
        self,
        input_features: mx.array,
        input_features_mask: Optional[mx.array] = None,
        audio_feature_lengths: Optional[mx.array] = None,
    ):
        return self.thinker.get_audio_features(
            input_features=input_features,
            feature_attention_mask=input_features_mask,
            audio_feature_lengths=audio_feature_lengths,
        )

    def get_image_features(
        self,
        pixel_values: mx.array,
        image_grid_thw: Optional[mx.array] = None,
    ):
        dtype = self.thinker.vision_tower.patch_embed.proj.weight.dtype
        pixel_values = pixel_values.astype(dtype)
        vision_output = self.thinker.vision_tower(pixel_values, image_grid_thw)
        if isinstance(vision_output, tuple):
            return vision_output[0]
        return vision_output

    @property
    def layers(self):
        return self.thinker.language_model.layers

    def extract_thinker_hidden_states(self, input_ids, target_layer_idx, **kwargs):
        embed_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k
            in [
                "pixel_values",
                "pixel_values_videos",
                "image_grid_thw",
                "video_grid_thw",
                "input_features",
                "feature_attention_mask",
                "audio_feature_lengths",
            ]
        }
        input_embedding_features = self.thinker.get_input_embeddings(
            input_ids, **embed_kwargs
        )
        inputs_embeds = input_embedding_features.inputs_embeds

        lm_kwargs = {
            k: v for k, v in kwargs.items() if k in ["image_grid_thw", "video_grid_thw"]
        }

        outputs = self.thinker.language_model(
            input_ids,
            inputs_embeds=inputs_embeds,
            output_hidden_states=True,
            **lm_kwargs,
        )

        hidden_states = outputs.hidden_states[target_layer_idx + 1]

        return hidden_states, inputs_embeds

    def _get_talker_user_parts(
        self,
        im_start_index: int,
        segment_end_index: int,
        multimodal_mask: mx.array,
        thinker_hidden: mx.array,
        thinker_embed: mx.array,
    ):
        seq_len = segment_end_index - im_start_index
        user_talker_part = mx.zeros(
            (1, seq_len, self.config.talker_config.text_config.hidden_size),
            dtype=thinker_embed.dtype,
        )
        user_mm_mask = multimodal_mask[:, im_start_index:segment_end_index]
        user_thinker_hidden_mm = thinker_hidden[:, im_start_index:segment_end_index]
        user_thinker_embed_seg = thinker_embed[:, im_start_index:segment_end_index]

        if mx.any(user_mm_mask):
            mm_indices = mx.array(
                np.where(np.array(mx.reshape(user_mm_mask, (-1,))))[0]
            )
            user_thinker_hidden_mm_flat = mx.reshape(
                user_thinker_hidden_mm, (-1, user_thinker_hidden_mm.shape[-1])
            )
            mm_hidden_flat = mx.take(user_thinker_hidden_mm_flat, mm_indices, axis=0)
            mm_hidden = self.talker.hidden_projection(mm_hidden_flat)
            user_talker_part_flat = mx.reshape(
                user_talker_part, (-1, user_talker_part.shape[-1])
            )
            user_talker_part_flat[mm_indices] = mm_hidden
            user_talker_part = mx.reshape(user_talker_part_flat, user_talker_part.shape)

        text_mask = ~user_mm_mask
        if mx.any(text_mask):
            text_indices = mx.array(np.where(np.array(mx.reshape(text_mask, (-1,))))[0])
            user_thinker_embed_flat = mx.reshape(
                user_thinker_embed_seg, (-1, user_thinker_embed_seg.shape[-1])
            )
            text_embed_flat = mx.take(user_thinker_embed_flat, text_indices, axis=0)
            user_text_hidden = self.talker.text_projection(text_embed_flat)
            user_talker_part_flat = mx.reshape(
                user_talker_part, (-1, user_talker_part.shape[-1])
            )
            user_talker_part_flat[text_indices] = user_text_hidden
            user_talker_part = mx.reshape(user_talker_part_flat, user_talker_part.shape)

        return user_talker_part

    def _get_talker_assistant_parts(
        self,
        im_start_index: int,
        segment_end_index: int,
        speaker_id: int,
        thinker_embed: mx.array,
        tts_pad_embed: mx.array,
        tts_bos_embed: mx.array,
        tts_eos_embed: mx.array,
    ):
        assistant_hidden = self.talker.text_projection(
            thinker_embed[:, im_start_index:segment_end_index]
        )
        assistant_text_hidden = mx.concatenate(
            (
                assistant_hidden[:, :3],
                mx.broadcast_to(tts_pad_embed, (1, 4, tts_pad_embed.shape[-1])),
                tts_bos_embed,
                assistant_hidden[:, 3:4],
            ),
            axis=1,
        )
        codec_special_tokens = mx.array(
            [
                [
                    self.config.talker_config.codec_nothink_id,
                    self.config.talker_config.codec_think_bos_id,
                    self.config.talker_config.codec_think_eos_id,
                    speaker_id,
                    self.config.talker_config.codec_pad_id,
                    self.config.talker_config.codec_bos_id,
                ]
            ],
            dtype=mx.int32,
        )
        assistant_codec_hidden = mx.concatenate(
            (
                mx.zeros(
                    (1, 3, self.config.talker_config.text_config.hidden_size),
                    dtype=thinker_embed.dtype,
                ),
                self.talker.model.codec_embedding(codec_special_tokens),
            ),
            axis=1,
        )
        trailing_text_hidden = mx.concatenate(
            (
                assistant_hidden[:, 4:],
                tts_eos_embed,
            ),
            axis=1,
        )
        input_embeds = assistant_text_hidden + assistant_codec_hidden
        input_ids = mx.full(
            (1, assistant_text_hidden.shape[1]),
            self.config.tts_pad_token_id,
            dtype=mx.int32,
        )
        return input_embeds, input_ids, trailing_text_hidden

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        pixel_values_videos: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):
        return self.thinker(
            input_ids=input_ids,
            pixel_values=pixel_values,
            pixel_values_videos=pixel_values_videos,
            mask=mask,
            cache=cache,
            **kwargs,
        )

    def generate(
        self,
        input_ids: mx.array,
        speaker: str = "Ethan",
        use_audio_in_video: bool = False,
        return_audio: Optional[bool] = None,
        thinker_max_new_tokens: int = 1024,
        thinker_eos_token_id: int = 151645,
        talker_max_new_tokens: int = 4096,
        talker_do_sample: bool = True,
        talker_top_k: int = 50,
        talker_top_p: float = 1.0,
        talker_temperature: float = 0.9,
        talker_repetition_penalty: float = 1.05,
        **kwargs,
    ):
        if return_audio and not self.has_talker:
            raise ValueError(
                "Cannot use talker when talker module not initialized. Use `enable_talker` method or set enable_audio_output in config to enable talker."
            )
        if return_audio is None:
            return_audio = self.has_talker

        if not return_audio:
            from mlx_vlm.generate import generate_step

            thinker_kwargs = {
                "max_tokens": thinker_max_new_tokens,
                "eos_tokens": [thinker_eos_token_id],
            }
            for key, value in kwargs.items():
                if key.startswith("thinker_"):
                    thinker_kwargs[key[len("thinker_") :]] = value
                elif key in (
                    "input_features",
                    "feature_attention_mask",
                    "audio_feature_lengths",
                    "pixel_values",
                    "pixel_values_videos",
                    "image_grid_thw",
                    "video_grid_thw",
                ):
                    thinker_kwargs[key] = value

            generator = generate_step(
                input_ids,
                self.thinker,
                thinker_kwargs.get("pixel_values"),
                kwargs.get("mask"),
                **{
                    k: v
                    for k, v in thinker_kwargs.items()
                    if k not in ("pixel_values", "mask")
                },
            )
            sequences = [input_ids]
            for token, _ in generator:
                sequences.append(mx.array([[token]]))
                if token == thinker_eos_token_id:
                    break
            thinker_result = type(
                "obj",
                (object,),
                {
                    "sequences": mx.concatenate(sequences, axis=1),
                    "hidden_states": None,
                },
            )()
            return thinker_result, None

        if input_ids.shape[0] != 1:
            raise NotImplementedError(
                "Qwen3-Omni currently does not support batched inference with audio output"
            )

        speaker_id = self.config.talker_config.speaker_id.get(speaker.lower())
        if speaker_id is None:
            raise NotImplementedError(f"Speaker {speaker} not implemented")

        from mlx_vlm.generate import generate_step

        thinker_kwargs = {
            "max_tokens": thinker_max_new_tokens,
            "eos_tokens": [thinker_eos_token_id],
            "output_hidden_states": True,
        }
        for key, value in kwargs.items():
            if key.startswith("thinker_"):
                thinker_kwargs[key[len("thinker_") :]] = value
            elif key in (
                "input_features",
                "feature_attention_mask",
                "audio_feature_lengths",
                "pixel_values",
                "pixel_values_videos",
                "image_grid_thw",
                "video_grid_thw",
            ):
                thinker_kwargs[key] = value

        generator = generate_step(
            input_ids,
            self.thinker,
            thinker_kwargs.get("pixel_values"),
            kwargs.get("mask"),
            **{
                k: v
                for k, v in thinker_kwargs.items()
                if k not in ("pixel_values", "mask", "output_hidden_states")
            },
        )
        sequences = [input_ids]
        hidden_states_list = []
        for token, _ in generator:
            sequences.append(mx.array([[token]]))
            if token == thinker_eos_token_id:
                break

        thinker_result_sequences = mx.concatenate(sequences, axis=1)

        thinker_hidden_all, thinker_embed_all = self.extract_thinker_hidden_states(
            thinker_result_sequences,
            target_layer_idx=self.config.talker_config.accept_hidden_layer,
            **kwargs,
        )

        im_start_indexes = mx.concatenate(
            (
                mx.array(
                    np.where(np.array(input_ids[0] == self.config.im_start_token_id))[0]
                ),
                mx.array([thinker_result_sequences.shape[-1]], dtype=mx.int32),
            ),
            axis=0,
        )
        multimodal_mask = (
            (thinker_result_sequences == self.config.thinker_config.audio_token_id)
            | (thinker_result_sequences == self.config.thinker_config.image_token_id)
            | (thinker_result_sequences == self.config.thinker_config.video_token_id)
        )

        talker_special_tokens = mx.array(
            [
                [
                    self.config.tts_bos_token_id,
                    self.config.tts_eos_token_id,
                    self.config.tts_pad_token_id,
                ]
            ],
            dtype=input_ids.dtype,
        )
        talker_special_embeds = self.thinker.language_model.model.embed_tokens(
            talker_special_tokens
        )
        talker_special_embeds_proj = self.talker.text_projection(talker_special_embeds)
        tts_bos_embed = talker_special_embeds_proj[:, 0:1]
        tts_eos_embed = talker_special_embeds_proj[:, 1:2]
        tts_pad_embed = talker_special_embeds_proj[:, 2:3]

        talker_input_embeds = []
        talker_input_ids = []

        for i in range(len(im_start_indexes) - 1):
            im_start_index = int(im_start_indexes[i])
            segment_end_index = int(im_start_indexes[i + 1])
            role_token = int(input_ids[0, im_start_index + 1])

            if role_token == self.config.system_token_id:
                continue
            elif role_token == self.config.user_token_id:
                talker_user_part = self._get_talker_user_parts(
                    im_start_index,
                    segment_end_index,
                    multimodal_mask,
                    thinker_hidden_all,
                    thinker_embed_all,
                )
                talker_input_embeds.append(talker_user_part)
                talker_input_ids.append(
                    thinker_result_sequences[:, im_start_index:segment_end_index]
                )
            elif (
                role_token == self.config.assistant_token_id
                and i == len(im_start_indexes) - 2
            ):
                talker_assistant_embeds, talker_assistant_ids, trailing_text_hidden = (
                    self._get_talker_assistant_parts(
                        im_start_index,
                        segment_end_index,
                        speaker_id,
                        thinker_embed_all,
                        tts_pad_embed,
                        tts_bos_embed,
                        tts_eos_embed,
                    )
                )
                talker_input_embeds.append(talker_assistant_embeds)
                talker_input_ids.append(talker_assistant_ids)
            elif (
                role_token == self.config.assistant_token_id
                and i != len(im_start_indexes) - 2
            ):
                continue
            else:
                raise AssertionError(
                    "Expect role id after <|im_start|> (assistant, user, system)"
                )

        if len(talker_input_embeds) == 0:
            return (
                type(
                    "obj",
                    (object,),
                    {
                        "sequences": thinker_result_sequences,
                        "hidden_states": None,
                    },
                )(),
                None,
            )

        talker_input_embed = mx.concatenate(talker_input_embeds, axis=1)
        talker_input_id = mx.concatenate(talker_input_ids, axis=1)

        talker_result = self.talker.generate(
            inputs_embeds=talker_input_embed,
            trailing_text_hidden=trailing_text_hidden,
            tts_pad_embed=tts_pad_embed,
            talker_input_ids=talker_input_id,
            max_new_tokens=talker_max_new_tokens,
            temperature=talker_temperature,
            top_p=talker_top_p,
        )

        valid_codes = [
            hid[-1] for hid in talker_result.hidden_states if hid[-1] is not None
        ]
        if not valid_codes:
            talker_wavs = mx.zeros((1, 1, 1000))
        else:
            talker_codes = mx.stack(valid_codes, axis=1).transpose(0, 2, 1)
            talker_wavs = self.code2wav.chunked_decode(
                talker_codes, chunk_size=300, left_context_size=25
            )

        thinker_result = type(
            "obj",
            (object,),
            {
                "sequences": thinker_result_sequences,
                "hidden_states": None,
            },
        )()

        return thinker_result, talker_wavs.astype(mx.float32)

    def generate_stream(
        self,
        input_ids: mx.array,
        speaker: str = "Ethan",
        thinker_max_new_tokens: int = 1024,
        thinker_eos_token_id: int = 151645,
        talker_max_new_tokens: int = 4096,
        talker_top_p: float = 1.0,
        talker_temperature: float = 0.9,
        chunk_size: int = 300,
        left_context_size: int = 25,
        **kwargs,
    ):
        if not self.has_talker:
            raise ValueError("Cannot stream audio without talker module")
        if input_ids.shape[0] != 1:
            raise NotImplementedError("Streaming does not support batched inference")

        speaker_id = self.config.talker_config.speaker_id.get(speaker.lower())
        if speaker_id is None:
            raise NotImplementedError(f"Speaker {speaker} not implemented")

        from mlx_vlm.generate import generate_step

        thinker_kwargs = {
            "max_tokens": thinker_max_new_tokens,
            "eos_tokens": [thinker_eos_token_id],
        }
        for key, value in kwargs.items():
            if key.startswith("thinker_"):
                thinker_kwargs[key[len("thinker_") :]] = value
            elif key in (
                "input_features",
                "feature_attention_mask",
                "audio_feature_lengths",
                "pixel_values",
                "pixel_values_videos",
                "image_grid_thw",
                "video_grid_thw",
            ):
                thinker_kwargs[key] = value

        generator = generate_step(
            input_ids,
            self.thinker,
            thinker_kwargs.get("pixel_values"),
            kwargs.get("mask"),
            **{
                k: v
                for k, v in thinker_kwargs.items()
                if k not in ("pixel_values", "mask")
            },
        )
        sequences = [input_ids]
        for token, _ in generator:
            sequences.append(mx.array([[token]]))
            if token == thinker_eos_token_id:
                break

        thinker_result_sequences = mx.concatenate(sequences, axis=1)
        thinker_hidden_all, thinker_embed_all = self.extract_thinker_hidden_states(
            thinker_result_sequences,
            target_layer_idx=self.config.talker_config.accept_hidden_layer,
            **kwargs,
        )

        im_start_indexes = mx.concatenate(
            (
                mx.array(
                    np.where(np.array(input_ids[0] == self.config.im_start_token_id))[0]
                ),
                mx.array([thinker_result_sequences.shape[-1]], dtype=mx.int32),
            ),
            axis=0,
        )
        multimodal_mask = (
            (thinker_result_sequences == self.config.thinker_config.audio_token_id)
            | (thinker_result_sequences == self.config.thinker_config.image_token_id)
            | (thinker_result_sequences == self.config.thinker_config.video_token_id)
        )

        talker_special_tokens = mx.array(
            [
                [
                    self.config.tts_bos_token_id,
                    self.config.tts_eos_token_id,
                    self.config.tts_pad_token_id,
                ]
            ],
            dtype=input_ids.dtype,
        )
        talker_special_embeds = self.thinker.language_model.model.embed_tokens(
            talker_special_tokens
        )
        talker_special_embeds_proj = self.talker.text_projection(talker_special_embeds)
        tts_bos_embed, tts_eos_embed, tts_pad_embed = (
            talker_special_embeds_proj[:, 0:1],
            talker_special_embeds_proj[:, 1:2],
            talker_special_embeds_proj[:, 2:3],
        )

        talker_input_embeds, talker_input_ids = [], []
        trailing_text_hidden = None

        for i in range(len(im_start_indexes) - 1):
            im_start_index, segment_end_index = int(im_start_indexes[i]), int(
                im_start_indexes[i + 1]
            )
            role_token = int(input_ids[0, im_start_index + 1])

            if role_token == self.config.system_token_id:
                continue
            elif role_token == self.config.user_token_id:
                talker_input_embeds.append(
                    self._get_talker_user_parts(
                        im_start_index,
                        segment_end_index,
                        multimodal_mask,
                        thinker_hidden_all,
                        thinker_embed_all,
                    )
                )
                talker_input_ids.append(
                    thinker_result_sequences[:, im_start_index:segment_end_index]
                )
            elif (
                role_token == self.config.assistant_token_id
                and i == len(im_start_indexes) - 2
            ):
                talker_assistant_embeds, talker_assistant_ids, trailing_text_hidden = (
                    self._get_talker_assistant_parts(
                        im_start_index,
                        segment_end_index,
                        speaker_id,
                        thinker_embed_all,
                        tts_pad_embed,
                        tts_bos_embed,
                        tts_eos_embed,
                    )
                )
                talker_input_embeds.append(talker_assistant_embeds)
                talker_input_ids.append(talker_assistant_ids)

        if not talker_input_embeds:
            return

        talker_input_embed = mx.concatenate(talker_input_embeds, axis=1)
        talker_input_id = mx.concatenate(talker_input_ids, axis=1)

        generated_tokens = thinker_result_sequences[0, input_ids.shape[1] :].tolist()
        yield ("text", generated_tokens)

        codes_list = []
        decoded_len = 0

        for residual_codes in self.talker.generate_stream(
            inputs_embeds=talker_input_embed,
            trailing_text_hidden=trailing_text_hidden,
            tts_pad_embed=tts_pad_embed,
            talker_input_ids=talker_input_id,
            max_new_tokens=talker_max_new_tokens,
            temperature=talker_temperature,
            top_p=talker_top_p,
        ):
            codes_list.append(residual_codes)
            if len(codes_list) >= chunk_size:
                codes_buffer = mx.stack(codes_list, axis=1).transpose(0, 2, 1)
                wav_chunk, decoded_len = self.code2wav.stream_decode(
                    codes_buffer, chunk_size, left_context_size, decoded_len
                )
                if wav_chunk is not None:
                    mx.eval(wav_chunk)
                    yield ("audio", wav_chunk.astype(mx.float32))

        if codes_list:
            codes_buffer = mx.stack(codes_list, axis=1).transpose(0, 2, 1)
            wav_chunk = self.code2wav.flush_decode(
                codes_buffer, left_context_size, decoded_len
            )
            if wav_chunk is not None:
                mx.eval(wav_chunk)
                yield ("audio", wav_chunk.astype(mx.float32))



"""
MLX-based Molmo Processor.

This module provides an MLX-native processor for Molmo models that doesn't
require torch, torchvision, or tensorflow.
"""

import json
from pathlib import Path
from typing import List, Optional, Tuple, Union

import mlx.core as mx
import numpy as np
from PIL import Image
from transformers import AutoTokenizer
from transformers.feature_extraction_utils import BatchFeature
from transformers.image_processing_utils import BaseImageProcessor
from transformers.image_utils import ImageInput, make_list_of_images
from transformers.processing_utils import ProcessorMixin

# CLIP normalization constants
OPENAI_CLIP_MEAN = (0.48145466, 0.4578275, 0.40821073)
OPENAI_CLIP_STD = (0.26862954, 0.26130258, 0.27577711)


def pad_to_bounding_box(
    image: np.ndarray,
    offset_height: int,
    offset_width: int,
    target_height: int,
    target_width: int,
    value: int = 0,
) -> np.ndarray:
    """Pad image to target bounding box."""
    height, width = image.shape[:2]
    after_padding_width = target_width - offset_width - width
    after_padding_height = target_height - offset_height - height
    if image.ndim == 3:
        padding = [
            [offset_height, after_padding_height],
            [offset_width, after_padding_width],
            [0, 0],
        ]
    else:
        padding = [
            [offset_height, after_padding_height],
            [offset_width, after_padding_width],
        ]
    return np.pad(image, padding, constant_values=value)


def normalize_image(
    image: np.ndarray, offset: Tuple[float, ...], scale: Tuple[float, ...]
) -> np.ndarray:
    """Normalize image with mean and std."""
    image = image.astype(np.float32)
    image -= np.array(offset, dtype=np.float32)[None, None, :]
    image /= np.array(scale, dtype=np.float32)[None, None, :]
    return image


def resize_and_pad(
    image: np.ndarray,
    desired_output_size: Tuple[int, int],
    pad_value: float = 0,
    normalize: bool = True,
    image_mean: Tuple[float, ...] = OPENAI_CLIP_MEAN,
    image_std: Tuple[float, ...] = OPENAI_CLIP_STD,
) -> Tuple[np.ndarray, np.ndarray]:
    """Resize and pad image using PIL (no torch/tensorflow)."""
    desired_height, desired_width = desired_output_size
    height, width = image.shape[:2]

    # Calculate scale
    image_scale_y = np.float32(desired_height) / np.float32(height)
    image_scale_x = np.float32(desired_width) / np.float32(width)
    image_scale = min(image_scale_x, image_scale_y)
    scaled_height = int(np.float32(height) * image_scale)
    scaled_width = int(np.float32(width) * image_scale)

    # Use PIL for resizing (bilinear interpolation)
    pil_image = Image.fromarray(
        (image * 255).astype(np.uint8) if image.max() <= 1.0 else image.astype(np.uint8)
    )
    pil_image = pil_image.resize(
        (scaled_width, scaled_height), Image.Resampling.BILINEAR
    )
    image = np.array(pil_image).astype(np.float32) / 255.0
    image = np.clip(image, 0.0, 1.0)

    # Pad to desired size
    top_pad = (desired_height - scaled_height) // 2
    left_pad = (desired_width - scaled_width) // 2
    padding = [
        [top_pad, desired_height - scaled_height - top_pad],
        [left_pad, desired_width - scaled_width - left_pad],
        [0, 0],
    ]
    image_mask = np.pad(np.ones_like(image[:, :, 0], dtype=bool), padding[:2])
    image = np.pad(image, padding, constant_values=pad_value)

    if normalize:
        image = normalize_image(image, offset=image_mean, scale=image_std)

    return image, image_mask


def select_tiling(
    h: int, w: int, patch_size: int, max_num_patches: int
) -> Tuple[int, int]:
    """Select best tiling for image."""
    tilings = []
    for i in range(1, max_num_patches + 1):
        for j in range(1, max_num_patches + 1):
            if i * j <= max_num_patches:
                tilings.append((i, j))
    tilings.sort(key=lambda x: (x[0] * x[1], x[0]))
    candidate_tilings = np.array(tilings, dtype=np.int32)
    candidate_resolutions = candidate_tilings * patch_size

    original_size = np.array([h, w], dtype=np.float32)
    required_scale_d = candidate_resolutions.astype(np.float32) / original_size
    required_scale = np.min(required_scale_d, axis=-1, keepdims=True)

    if np.all(required_scale < 1):
        ix = np.argmax(required_scale)
    else:
        required_scale = np.where(required_scale < 1.0, 10e9, required_scale)
        ix = np.argmin(required_scale)

    return tuple(candidate_tilings[ix])


def rearrange_patches(
    patches: np.ndarray, dh: int, dw: int, h: int, w: int
) -> np.ndarray:
    """Rearrange patches: 'p (h dh) (w dw) c -> p (h w) (dh dw c)'"""
    p, H, W, c = patches.shape
    patches = patches.reshape(p, h, dh, w, dw, c)
    patches = patches.transpose(0, 1, 3, 2, 4, 5)
    patches = patches.reshape(p, h * w, dh * dw * c)
    return patches


def rearrange_mask(mask: np.ndarray, dh: int, dw: int, h: int, w: int) -> np.ndarray:
    """Rearrange mask: 'p (h dh) (w dw) -> p (h w) (dh dw)'"""
    p, H, W = mask.shape
    mask = mask.reshape(p, h, dh, w, dw)
    mask = mask.transpose(0, 1, 3, 2, 4)
    mask = mask.reshape(p, h * w, dh * dw)
    return mask


def rearrange_global(image: np.ndarray, dh: int, dw: int, h: int, w: int) -> np.ndarray:
    """Rearrange global image: '(h dh) (w dw) c -> (h w) (dh dw c)'"""
    H, W, c = image.shape
    image = image.reshape(h, dh, w, dw, c)
    image = image.transpose(0, 2, 1, 3, 4)
    image = image.reshape(h * w, dh * dw * c)
    return image


class MolmoImageProcessor(BaseImageProcessor):
    """MLX-based image processor for Molmo."""

    model_input_names = ["images", "image_input_idx", "image_masks"]

    def __init__(
        self,
        max_crops: int = 12,
        overlap_margins: List[int] = None,
        base_image_input_size: List[int] = None,
        image_token_length_w: int = 12,
        image_token_length_h: int = 12,
        image_patch_size: int = 14,
        image_padding_mask: bool = True,
        do_normalize: bool = True,
        image_mean: Optional[List[float]] = None,
        image_std: Optional[List[float]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.max_crops = max_crops
        self.overlap_margins = overlap_margins or [4, 4]
        self.base_image_input_size = base_image_input_size or [336, 336]
        self.image_token_length_w = image_token_length_w
        self.image_token_length_h = image_token_length_h
        self.image_patch_size = image_patch_size
        self.image_padding_mask = image_padding_mask
        self.do_normalize = do_normalize
        self.image_mean = tuple(image_mean) if image_mean else OPENAI_CLIP_MEAN
        self.image_std = tuple(image_std) if image_std else OPENAI_CLIP_STD

    def image_to_patches_and_tokens(
        self,
        image: np.ndarray,
        image_patch_token_id: int,
        image_col_token_id: int,
        image_start_token_id: int,
        image_end_token_id: int,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Convert image to patches and tokens."""
        base_image_input_size = self.base_image_input_size
        base_image_input_d = self.image_patch_size
        tokens_per_image = self.image_token_length_w * self.image_token_length_h
        image_base_patch_w = base_image_input_size[1] // base_image_input_d
        image_base_patch_h = base_image_input_size[0] // base_image_input_d

        original_image_h, original_image_w = image.shape[:2]
        crop_size = base_image_input_size[0]

        left_margin, right_margin = self.overlap_margins
        total_margin_pixels = base_image_input_d * (right_margin + left_margin)
        crop_patches = base_image_input_size[0] // base_image_input_d
        crop_window_patches = crop_patches - (right_margin + left_margin)
        crop_window_size = crop_window_patches * base_image_input_d

        tiling = select_tiling(
            original_image_h - total_margin_pixels,
            original_image_w - total_margin_pixels,
            crop_window_size,
            self.max_crops,
        )

        src, img_mask = resize_and_pad(
            image,
            [
                tiling[0] * crop_window_size + total_margin_pixels,
                tiling[1] * crop_window_size + total_margin_pixels,
            ],
            image_mean=self.image_mean,
            image_std=self.image_std,
        )

        patches_arr = []
        mask_arr = []
        patch_ordering_arr = []

        on = 0
        for i in range(tiling[0]):
            y0 = i * crop_window_size
            crop_y0 = 0 if i == 0 else left_margin // 2

            crop_h = image_base_patch_h - (right_margin + left_margin)
            if i == 0:
                crop_h += left_margin
            if i == (tiling[0] - 1):
                crop_h += right_margin

            for j in range(tiling[1]):
                x0 = j * crop_window_size
                crop_x0 = 0 if j == 0 else left_margin // 2

                crop_w = image_base_patch_w - (right_margin + left_margin)
                if j == 0:
                    crop_w += left_margin
                if j == (tiling[1] - 1):
                    crop_w += right_margin

                pooled_w = (crop_w + 1) // 2
                pooled_h = (crop_h + 1) // 2

                ordering = np.reshape(
                    np.arange(on, on + pooled_h * pooled_w, dtype=np.int32),
                    (pooled_h, pooled_w, 1),
                )
                patch_ordering_arr.append(
                    pad_to_bounding_box(
                        ordering,
                        crop_y0,
                        crop_x0,
                        self.image_token_length_h,
                        self.image_token_length_w,
                        value=-1,
                    )[:, :, 0]
                )
                patches_arr.append(src[y0 : y0 + crop_size, x0 : x0 + crop_size])
                mask_arr.append(img_mask[y0 : y0 + crop_size, x0 : x0 + crop_size])

                on += pooled_h * pooled_w

        patches = np.stack(patches_arr)
        patch_ordering = np.stack(patch_ordering_arr)
        img_mask = np.stack(mask_arr)

        # Rearrange patches
        patches = rearrange_patches(
            patches,
            base_image_input_d,
            base_image_input_d,
            image_base_patch_h,
            image_base_patch_w,
        )
        img_mask = rearrange_mask(
            img_mask,
            base_image_input_d,
            base_image_input_d,
            image_base_patch_h,
            image_base_patch_w,
        )

        img_mask = img_mask.astype(np.float32).mean(axis=-1)
        patch_ordering = np.reshape(patch_ordering, [-1])
        valid = patch_ordering >= 0

        # Transpose order
        patch_ordering_rh = np.reshape(
            patch_ordering,
            [
                tiling[0],
                tiling[1],
                self.image_token_length_h,
                self.image_token_length_w,
            ],
        )
        patch_ordering_rh = np.transpose(patch_ordering_rh, [0, 2, 1, 3])
        patch_ordering_rh = np.reshape(patch_ordering_rh, [-1])

        patch_ordering[valid] = patch_ordering_rh[patch_ordering_rh >= 0]

        # Build output tokens
        h = tiling[0] * crop_window_patches + (right_margin + left_margin)
        w = tiling[1] * crop_window_patches + (right_margin + left_margin)
        per_row = np.full(((w + 1) // 2,), image_patch_token_id)
        per_row = np.concatenate([per_row, [image_col_token_id]], 0)

        joint = np.tile(per_row, [(h + 1) // 2])
        joint = [[image_start_token_id], joint, [image_end_token_id]]

        # Global image
        resized, _ = resize_and_pad(
            image,
            base_image_input_size,
            image_mean=self.image_mean,
            image_std=self.image_std,
        )
        resized = rearrange_global(
            resized,
            base_image_input_d,
            base_image_input_d,
            image_base_patch_h,
            image_base_patch_w,
        )
        patches = np.concatenate([np.expand_dims(resized, 0), patches], 0)

        patch_ordering = np.where(
            patch_ordering >= 0, patch_ordering + tokens_per_image, -1
        )
        patch_ordering = np.concatenate(
            [np.arange(0, tokens_per_image), patch_ordering], 0
        )

        per_row = np.full((self.image_token_length_w,), image_patch_token_id)
        per_row = np.concatenate([per_row, [image_col_token_id]], 0)
        extra_tokens = np.tile(per_row, [self.image_token_length_h])
        joint = [
            [image_start_token_id],
            extra_tokens,
            [image_end_token_id],
        ] + joint

        joint = np.concatenate(joint, 0)
        img_mask = np.pad(img_mask, [[0, 1], [0, 0]], constant_values=-1)

        return patches, joint, patch_ordering, img_mask

    def build_image_input_idx(
        self,
        image_tokens: np.ndarray,
        patch_order: np.ndarray,
        image_patch_token_id: int,
    ) -> np.ndarray:
        """Build image input indices."""
        tokens_per_image = self.image_token_length_w * self.image_token_length_h

        image_input_idx = image_tokens == image_patch_token_id
        image_input_idx = np.nonzero(image_input_idx)[0].astype(np.int32)

        if patch_order is not None:
            n_tokens = image_input_idx.shape[0]
            patch_order = np.reshape(patch_order, [-1])

            valid = patch_order >= 0
            n_valid_patches = valid.sum()

            sorted_patch_ixs = np.zeros([n_tokens], np.int32)
            sorted_patch_ixs[patch_order[valid]] = np.arange(
                n_valid_patches, dtype=np.int32
            )

            sorted_patch_ixs_ex = np.full(np.shape(patch_order), -1)
            sorted_patch_ixs_ex[valid] = sorted_patch_ixs

            valid_int = (sorted_patch_ixs_ex >= 0).astype(np.int32)
            image_input_idx = image_input_idx[sorted_patch_ixs_ex * valid_int]
            image_input_idx = image_input_idx * valid_int - 100 * (1 - valid_int)
            image_input_idx = np.reshape(image_input_idx, [-1, tokens_per_image])

        return image_input_idx

    def preprocess(
        self,
        image: np.ndarray,
        image_patch_token_id: int,
        image_col_token_id: int,
        image_start_token_id: int,
        image_end_token_id: int,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Preprocess a single image."""
        crops, image_tokens, patch_ordering, img_mask = (
            self.image_to_patches_and_tokens(
                image,
                image_patch_token_id,
                image_col_token_id,
                image_start_token_id,
                image_end_token_id,
            )
        )
        patch_idx = self.build_image_input_idx(
            image_tokens, patch_ordering, image_patch_token_id
        )
        return crops, image_tokens, patch_idx, img_mask


class MolmoProcessor(ProcessorMixin):
    """MLX-based processor for Molmo."""

    attributes = ["image_processor", "tokenizer"]
    image_processor_class = "MolmoImageProcessor"
    tokenizer_class = "AutoTokenizer"

    def __init__(
        self,
        image_processor=None,
        tokenizer=None,
        chat_template=None,
        **kwargs,
    ):
        if image_processor is None:
            image_processor = MolmoImageProcessor()
        # Molmo uses these specific token names
        self.image_patch_token = "<im_patch>"
        self.image_col_token = "<im_col>"
        self.image_start_token = "<im_start>"
        self.image_end_token = "<im_end>"
        super().__init__(image_processor, tokenizer, chat_template=chat_template)

    def __call__(
        self,
        images: ImageInput = None,
        text: Union[str, List[str]] = None,
        **kwargs,
    ) -> BatchFeature:
        """Process images and text for Molmo."""
        if images is None and text is None:
            raise ValueError("You must provide either images or text.")

        # Get token IDs
        image_patch_token_id = self.tokenizer.convert_tokens_to_ids(
            self.image_patch_token
        )
        image_col_token_id = self.tokenizer.convert_tokens_to_ids(self.image_col_token)
        image_start_token_id = self.tokenizer.convert_tokens_to_ids(
            self.image_start_token
        )
        image_end_token_id = self.tokenizer.convert_tokens_to_ids(self.image_end_token)

        # Validate token IDs
        if image_patch_token_id is None:
            raise ValueError(
                f"Token '{self.image_patch_token}' not found in tokenizer vocabulary"
            )
        if image_col_token_id is None:
            raise ValueError(
                f"Token '{self.image_col_token}' not found in tokenizer vocabulary"
            )
        if image_start_token_id is None:
            raise ValueError(
                f"Token '{self.image_start_token}' not found in tokenizer vocabulary"
            )
        if image_end_token_id is None:
            raise ValueError(
                f"Token '{self.image_end_token}' not found in tokenizer vocabulary"
            )

        # Process images
        if images is not None:
            images = make_list_of_images(images)
            # Convert PIL images to numpy arrays
            np_images = []
            for img in images:
                if isinstance(img, Image.Image):
                    img = img.convert("RGB")
                    np_images.append(np.array(img).astype(np.float32) / 255.0)
                elif isinstance(img, np.ndarray):
                    if img.max() > 1.0:
                        img = img.astype(np.float32) / 255.0
                    np_images.append(img)
                else:
                    np_images.append(np.array(img).astype(np.float32) / 255.0)
            images = np_images

        # Tokenize text
        if text is not None:
            if isinstance(text, str):
                text = [text]
            tokens_list = [self.tokenizer.encode(t) for t in text]
        else:
            tokens_list = [[]]

        # Process each image with text
        if images is not None and len(images) > 0:
            all_crops = []
            all_image_idx = []
            all_masks = []
            all_input_ids = []

            for i, (img, tokens) in enumerate(
                zip(
                    images,
                    (
                        tokens_list
                        if len(tokens_list) == len(images)
                        else [tokens_list[0]] * len(images)
                    ),
                )
            ):
                crops, image_tokens, patch_idx, img_mask = (
                    self.image_processor.preprocess(
                        img,
                        image_patch_token_id,
                        image_col_token_id,
                        image_start_token_id,
                        image_end_token_id,
                    )
                )

                # Combine image tokens with text tokens
                combined_tokens = np.concatenate([image_tokens, np.array(tokens)])

                # Adjust patch_idx for the position in combined tokens
                all_crops.append(crops)
                all_image_idx.append(patch_idx)
                all_masks.append(img_mask)
                all_input_ids.append(combined_tokens)

            # Stack results
            pixel_values = mx.array(
                np.concatenate(all_crops, axis=0).astype(np.float32)
            )
            image_input_idx = mx.array(
                np.concatenate(all_image_idx, axis=0).astype(np.int32)
            )
            image_masks = mx.array(np.concatenate(all_masks, axis=0).astype(np.float32))

            # Pad input_ids to same length
            max_len = max(len(ids) for ids in all_input_ids)
            pad_token_id = self.tokenizer.pad_token_id
            if pad_token_id is None:
                pad_token_id = self.tokenizer.eos_token_id or 0
            padded_ids = []
            for ids in all_input_ids:
                pad_len = max_len - len(ids)
                if pad_len > 0:
                    ids = np.pad(ids, (0, pad_len), constant_values=pad_token_id)
                padded_ids.append(ids.astype(np.int32))

            input_ids = mx.array(np.stack(padded_ids).astype(np.int32))

            return BatchFeature(
                data={
                    "input_ids": input_ids,
                    "pixel_values": pixel_values,
                    "image_input_idx": image_input_idx,
                    "image_masks": image_masks,
                }
            )
        else:
            # Text only
            max_len = max(len(t) for t in tokens_list)
            pad_token_id = self.tokenizer.pad_token_id
            if pad_token_id is None:
                pad_token_id = self.tokenizer.eos_token_id or 0
            padded = []
            for t in tokens_list:
                pad_len = max_len - len(t)
                if pad_len > 0:
                    t = t + [pad_token_id] * pad_len
                padded.append(t)

            return BatchFeature(data={"input_ids": mx.array(padded, dtype=mx.int32)})

    def batch_decode(self, *args, **kwargs):
        """Forward to tokenizer's batch_decode."""
        return self.tokenizer.batch_decode(*args, **kwargs)

    def decode(self, *args, **kwargs):
        """Forward to tokenizer's decode."""
        return self.tokenizer.decode(*args, **kwargs)

    def apply_chat_template(
        self,
        conversation,
        chat_template=None,
        add_generation_prompt=False,
        tokenize=False,
        **kwargs,
    ):
        """Apply chat template."""
        if chat_template is None:
            chat_template = self.chat_template
        if chat_template is None:
            chat_template = getattr(self.tokenizer, "chat_template", None)
        if chat_template is None:
            # Default Molmo chat template
            chat_template = (
                "{% for message in messages %}"
                "{% if message['role'] == 'user' %}"
                "User: {{ message['content'] }}\n"
                "{% elif message['role'] == 'assistant' %}"
                "Assistant: {{ message['content'] }}\n"
                "{% endif %}"
                "{% endfor %}"
                "{% if add_generation_prompt %}Assistant: {% endif %}"
            )

        from jinja2 import Environment

        # Use Environment with loopcontrols extension to support {% continue %} and {% break %}
        env = Environment(extensions=["jinja2.ext.loopcontrols"])
        template = env.from_string(chat_template)
        rendered = template.render(
            messages=conversation,
            add_generation_prompt=add_generation_prompt,
            **kwargs,
        )

        if tokenize:
            return self.tokenizer.encode(rendered)
        return rendered

    @property
    def model_input_names(self):
        """Get model input names."""
        return ["input_ids", "pixel_values", "image_input_idx", "image_masks"]

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, **kwargs):
        """Load processor from pretrained model."""
        from huggingface_hub import hf_hub_download

        kwargs.pop("trust_remote_code", None)

        model_path = Path(pretrained_model_name_or_path)
        is_local = model_path.exists() and model_path.is_dir()

        # Load tokenizer
        tokenizer = AutoTokenizer.from_pretrained(
            str(model_path) if is_local else pretrained_model_name_or_path,
            trust_remote_code=True,
            local_files_only=is_local,
        )

        # Load image processor config
        image_processor_config = {}
        try:
            if is_local:
                config_path = model_path / "preprocessor_config.json"
            else:
                config_path = Path(
                    hf_hub_download(
                        pretrained_model_name_or_path, "preprocessor_config.json"
                    )
                )
            if config_path.exists():
                with open(config_path, "r") as f:
                    config = json.load(f)
                for key in [
                    "max_crops",
                    "overlap_margins",
                    "base_image_input_size",
                    "image_token_length_w",
                    "image_token_length_h",
                    "image_patch_size",
                    "image_padding_mask",
                    "do_normalize",
                    "image_mean",
                    "image_std",
                ]:
                    if key in config:
                        image_processor_config[key] = config[key]
        except Exception:
            pass

        image_processor = MolmoImageProcessor(**image_processor_config)

        # Load chat template
        chat_template = getattr(tokenizer, "chat_template", None)
        if chat_template is None:
            try:
                if is_local:
                    jinja_path = model_path / "chat_template.jinja"
                else:
                    jinja_path = Path(
                        hf_hub_download(
                            pretrained_model_name_or_path, "chat_template.jinja"
                        )
                    )
                if jinja_path.exists():
                    chat_template = jinja_path.read_text(encoding="utf-8")
                    tokenizer.chat_template = chat_template
            except Exception:
                pass

        return cls(
            image_processor=image_processor,
            tokenizer=tokenizer,
            chat_template=chat_template,
        )


# Patch AutoProcessor for Molmo models
from transformers import AutoProcessor

_original_auto_processor_from_pretrained = AutoProcessor.from_pretrained


@classmethod
def _patched_auto_processor_from_pretrained(
    cls, pretrained_model_name_or_path, **kwargs
):
    """Patched from_pretrained that returns MolmoProcessor for molmo models."""
    from huggingface_hub import hf_hub_download

    model_path = Path(pretrained_model_name_or_path)
    is_local = model_path.exists() and model_path.is_dir()

    # Check if this is a molmo model
    is_molmo = False
    try:
        if is_local:
            config_path = model_path / "config.json"
        else:
            config_path = Path(
                hf_hub_download(pretrained_model_name_or_path, "config.json")
            )
        with open(config_path, "r") as f:
            config = json.load(f)
        model_type = config.get("model_type", "").lower()

        is_molmo = model_type == "molmo"
    except Exception:
        pass

    if is_molmo:
        return MolmoProcessor.from_pretrained(pretrained_model_name_or_path, **kwargs)

    return _original_auto_processor_from_pretrained.__func__(
        cls, pretrained_model_name_or_path, **kwargs
    )


AutoProcessor.from_pretrained = _patched_auto_processor_from_pretrained



"""DFNRope Vision Transformer for ERNIE 4.5 VL."""

from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from .config import VisionConfig


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return mx.concatenate([-x2, x1], axis=-1)


def apply_rotary_pos_emb_vision(tensor: mx.array, freqs: mx.array) -> mx.array:
    """Applies Rotary Position Embedding to the input tensors.

    Args:
        tensor: The input tensor.
        freqs: The frequencies used for the rotation.

    Returns:
        output: the tensor rotated using the Rotary Position Embedding.
    """
    orig_dtype = tensor.dtype
    tensor = tensor.astype(mx.float32)
    cos = mx.cos(freqs)
    sin = mx.sin(freqs)
    # cos [θ0,θ1,θ2......θd/2-1] -> cos_pos [θ0,θ0,θ1,θ1,θ2,θ2......θd/2-1,θd/2-1]
    cos = mx.expand_dims(cos, axis=1)
    cos = mx.tile(cos, (1, 1, 2))
    cos = mx.expand_dims(cos, axis=0)

    sin = mx.expand_dims(sin, axis=1)
    sin = mx.tile(sin, (1, 1, 2))
    sin = mx.expand_dims(sin, axis=0)

    output = tensor * cos + rotate_half(tensor) * sin
    return output.astype(orig_dtype)


class VisionRotaryEmbedding(nn.Module):
    """Rotary position embedding for vision transformer."""

    def __init__(self, dim: int, theta: float = 10000.0) -> None:
        super().__init__()
        self.dim = dim
        self.theta = theta

    def __call__(self, seqlen: int) -> mx.array:
        inv_freq = 1.0 / (
            self.theta ** (mx.arange(0, self.dim, 2, dtype=mx.float32) / self.dim)
        )
        if isinstance(seqlen, mx.array):
            seqlen = seqlen.item()
        seq = mx.arange(seqlen, dtype=inv_freq.dtype)
        freqs = mx.outer(seq, inv_freq)
        return freqs


class PatchEmbed(nn.Module):
    """Linear patch embedding for DFNRope Vision Transformer."""

    def __init__(
        self,
        patch_size: int = 14,
        in_channels: int = 3,
        embed_dim: int = 1152,
    ) -> None:
        super().__init__()
        self.patch_size = patch_size
        self.in_channels = in_channels
        self.embed_dim = embed_dim
        # Linear projection: in_channels * patch_size * patch_size -> embed_dim
        self.proj = nn.Linear(
            in_channels * patch_size * patch_size, embed_dim, bias=False
        )

    def __call__(self, hidden_states: mx.array) -> mx.array:
        """
        Args:
            hidden_states: Input tensor of shape [num_patches, in_channels * patch_size * patch_size]
        Returns:
            Output tensor of shape [num_patches, embed_dim]
        """
        target_dtype = self.proj.weight.dtype
        hidden_states = self.proj(hidden_states.astype(target_dtype))
        return hidden_states


class VisionMLP(nn.Module):
    """MLP for vision transformer block."""

    def __init__(
        self, dim: int, hidden_dim: int, hidden_act: str = "quick_gelu"
    ) -> None:
        super().__init__()
        self.fc1 = nn.Linear(dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, dim)
        self.hidden_act = hidden_act

    def __call__(self, x: mx.array) -> mx.array:
        x = self.fc1(x)
        if self.hidden_act == "quick_gelu":
            x = x * mx.sigmoid(1.702 * x)
        elif self.hidden_act == "gelu":
            x = nn.gelu(x)
        elif self.hidden_act == "silu":
            x = nn.silu(x)
        else:
            x = nn.gelu(x)
        return self.fc2(x)


class VisionAttention(nn.Module):
    """Multi-head attention for vision transformer."""

    def __init__(self, dim: int, num_heads: int = 16) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5
        self.qkv = nn.Linear(dim, dim * 3, bias=True)
        self.proj = nn.Linear(dim, dim)

    def __call__(
        self,
        x: mx.array,
        cu_seqlens: mx.array,
        rotary_pos_emb: Optional[mx.array] = None,
    ) -> mx.array:
        """Forward function for vision attention."""
        seq_length = x.shape[0]
        qkv = (
            self.qkv(x).reshape(seq_length, 3, self.num_heads, -1).transpose(1, 0, 2, 3)
        )
        q, k, v = mx.split(qkv, 3)

        q = apply_rotary_pos_emb_vision(mx.expand_dims(q, 0), rotary_pos_emb)[0]
        k = apply_rotary_pos_emb_vision(mx.expand_dims(k, 0), rotary_pos_emb)[0]

        q = q.transpose(0, 2, 1, 3)
        k = k.transpose(0, 2, 1, 3)
        v = v.transpose(0, 2, 1, 3)

        lengths = (cu_seqlens[1:] - cu_seqlens[:-1]).tolist()
        splits = [
            mx.split(tensor, [lengths[0], sum(lengths[:2])], axis=2)
            for tensor in (q, k, v)
        ]

        attn_outputs = []
        for q, k, v in zip(*splits):
            output = mx.fast.scaled_dot_product_attention(
                q, k, v, scale=self.scale, mask=None
            )
            attn_outputs.append(output)

        output = mx.concatenate(attn_outputs, axis=2)
        output = output.transpose(0, 2, 1, 3).reshape(seq_length, -1)
        return self.proj(output)


class DFNRopeVisionBlock(nn.Module):
    """DFNRope Vision Transformer block."""

    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(config.embed_dim, eps=config.layer_norm_eps)
        self.norm2 = nn.LayerNorm(config.embed_dim, eps=config.layer_norm_eps)

        mlp_hidden_dim = int(config.embed_dim * config.mlp_ratio)
        self.attn = VisionAttention(config.embed_dim, num_heads=config.num_heads)
        self.mlp = VisionMLP(
            dim=config.embed_dim,
            hidden_dim=mlp_hidden_dim,
            hidden_act=config.hidden_act,
        )

    def __call__(
        self, hidden_states: mx.array, cu_seqlens: mx.array, rotary_pos_emb: mx.array
    ) -> mx.array:
        hidden_states = hidden_states + self.attn(
            self.norm1(hidden_states),
            cu_seqlens=cu_seqlens,
            rotary_pos_emb=rotary_pos_emb,
        )
        hidden_states = hidden_states + self.mlp(self.norm2(hidden_states))
        return hidden_states


class VisionModel(nn.Module):
    """DFNRope Vision Transformer for ERNIE 4.5 VL."""

    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        self.spatial_merge_size = config.spatial_merge_size

        self.patch_embed = PatchEmbed(
            patch_size=config.patch_size,
            in_channels=config.in_channels,
            embed_dim=config.embed_dim,
        )

        head_dim = config.embed_dim // config.num_heads
        self.rotary_pos_emb = VisionRotaryEmbedding(head_dim // 2)

        self.blocks = [DFNRopeVisionBlock(config) for _ in range(config.depth)]
        self.ln = nn.LayerNorm(config.embed_dim, eps=config.layer_norm_eps)

    def rot_pos_emb(self, grid_thw: mx.array, num_pad: int = 0) -> mx.array:
        """Compute rotary position embedding for vision.

        Args:
            grid_thw: Grid dimensions [batch, 3] containing (t, h, w)
            num_pad: Number of padding tokens

        Returns:
            Rotary position embedding tensor
        """
        pos_ids = []
        grid_hw_array = np.array(grid_thw.tolist(), dtype=np.int64)

        for t, h, w in grid_hw_array:
            hpos_ids = np.arange(h).reshape(-1, 1)
            hpos_ids = np.tile(hpos_ids, (1, w))
            hpos_ids = hpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            hpos_ids = np.transpose(hpos_ids, (0, 2, 1, 3))
            hpos_ids = hpos_ids.flatten()

            wpos_ids = np.arange(w).reshape(1, -1)
            wpos_ids = np.tile(wpos_ids, (h, 1))
            wpos_ids = wpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            wpos_ids = np.transpose(wpos_ids, (0, 2, 1, 3))
            wpos_ids = wpos_ids.flatten()

            stacked_ids = np.stack([hpos_ids, wpos_ids], axis=-1)
            tiled_ids = np.tile(stacked_ids, (t, 1))
            pos_ids.append(tiled_ids)

        pos_ids = np.concatenate(pos_ids, axis=0)
        if num_pad > 0:
            pos_ids = np.concatenate(
                [pos_ids, np.zeros((num_pad, 2), dtype=pos_ids.dtype)], axis=0
            )

        max_grid_size = int(np.max(grid_hw_array[:, 1:]))
        rotary_pos_emb_full = self.rotary_pos_emb(max_grid_size)
        pos_ids_mx = mx.array(pos_ids, dtype=mx.int32)
        rotary_pos_emb = rotary_pos_emb_full[pos_ids_mx].reshape(pos_ids.shape[0], -1)

        return rotary_pos_emb

    def __call__(
        self,
        hidden_states: mx.array,
        grid_thw: mx.array,
        output_hidden_states: Optional[bool] = None,
        num_pad: int = 0,
    ) -> mx.array:
        """Forward pass through the vision model.

        Args:
            hidden_states: Input pixel values [num_patches, channels * patch_h * patch_w]
            grid_thw: Grid dimensions [batch, 3]
            output_hidden_states: Whether to output hidden states
            num_pad: Number of padding tokens

        Returns:
            Vision features
        """
        hidden_states = self.patch_embed(hidden_states)
        rotary_pos_emb = self.rot_pos_emb(grid_thw, num_pad=num_pad)

        # Compute cumulative sequence lengths
        cu_seqlens = mx.zeros(1, dtype=mx.int32)
        for i in range(grid_thw.shape[0]):
            t, h, w = grid_thw[i].tolist()
            seq_len = t * h * w
            cu_seqlens = mx.concatenate([cu_seqlens, cu_seqlens[-1:] + seq_len])

        if num_pad > 0:
            cu_seqlens = mx.concatenate([cu_seqlens, cu_seqlens[-1:] + num_pad])

        encoder_states = (hidden_states,) if output_hidden_states else None

        for blk in self.blocks:
            hidden_states = blk(
                hidden_states,
                cu_seqlens=cu_seqlens,
                rotary_pos_emb=rotary_pos_emb,
            )
            if output_hidden_states:
                encoder_states = encoder_states + (hidden_states,)

        hidden_states = self.ln(hidden_states)
        return hidden_states

    def sanitize(self, weights):
        """Sanitize weights for loading."""
        sanitized_weights = {}
        for k, v in weights.items():
            if "position_ids" in k:
                continue
            sanitized_weights[k] = v
        return sanitized_weights



"""ERNIE 4.5 VL MoE model for MLX."""

from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from transformers import AutoImageProcessor, AutoProcessor, AutoTokenizer

from ..base import InputEmbeddingsFeatures
from .config import ModelConfig
from .language import LanguageModel
from .processor import Ernie4_5_VLProcessor, Ernie4_5_VLTokenizer, ImageProcessor
from .vision import VisionModel

# Register custom processor classes for ernie4_5_moe_vl model type
MODEL_TYPE = "ernie4_5_moe_vl"
try:
    AutoImageProcessor.register(MODEL_TYPE, slow_image_processor_class=ImageProcessor)
    AutoTokenizer.register(MODEL_TYPE, slow_tokenizer_class=Ernie4_5_VLTokenizer)
    AutoProcessor.register(MODEL_TYPE, Ernie4_5_VLProcessor)
except Exception:
    pass  # Already registered or registration not needed


class TokenType:
    """Token type definition."""

    text = 0
    image = 1
    video = 2


class VariableResolutionResamplerModel(nn.Module):
    """Compresses vision features using spatial and temporal convolutions."""

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        spatial_conv_size: int,
        temporal_conv_size: int,
        config: ModelConfig,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.config = config
        self.spatial_conv_size = spatial_conv_size
        self.temporal_conv_size = temporal_conv_size
        self.use_temporal_conv = config.use_temporal_conv

        self.spatial_dim = in_dim * spatial_conv_size * spatial_conv_size
        self.temporal_dim = (
            in_dim * spatial_conv_size * spatial_conv_size * temporal_conv_size
        )

        self.spatial_linear = nn.Sequential(
            nn.Linear(self.spatial_dim, self.spatial_dim),
            nn.GELU(),
            nn.Linear(self.spatial_dim, self.spatial_dim),
            nn.LayerNorm(self.spatial_dim, eps=1e-6),
        )

        if self.use_temporal_conv:
            self.temporal_linear = nn.Sequential(
                nn.Linear(self.temporal_dim, self.spatial_dim),
                nn.GELU(),
                nn.Linear(self.spatial_dim, self.spatial_dim),
                nn.LayerNorm(self.spatial_dim, eps=1e-6),
            )

        self.mlp = nn.Linear(self.spatial_dim, out_dim)
        self.after_norm = nn.RMSNorm(out_dim)

    def spatial_conv_reshape(self, x: mx.array) -> mx.array:
        S, C = x.shape
        x = x.reshape(-1, C * (self.spatial_conv_size**2))
        return x

    def __call__(
        self,
        x: mx.array,
        grid_thw: mx.array,
    ) -> mx.array:
        def fwd_spatial(x):
            x = self.spatial_conv_reshape(x)
            x = self.spatial_linear(x)
            return x

        def fwd_placeholder(x, grid_thw):
            grid_thw_np = np.array(grid_thw.tolist(), dtype=np.int64)
            grid_t = grid_thw_np[:, 0]
            grid_hw = grid_thw_np[:, 1:]
            grid_hw_after_conv = grid_hw.prod(-1) // (self.spatial_conv_size**2)

            tokens_per_img_or_vid = grid_thw_np.prod(-1) // (self.spatial_conv_size**2)
            batch_offset = np.empty(tokens_per_img_or_vid.size, dtype=np.int64)
            batch_offset[0] = 0
            batch_offset[1:] = tokens_per_img_or_vid.cumsum()[:-1]

            assert (
                self.temporal_conv_size == 2
            ), f"Hard Code: temporal_conv_size==2, got: {self.temporal_conv_size}"

            slice_offsets = []
            for temporal_size, spatial_size, b_offset in zip(
                grid_t, grid_hw_after_conv, batch_offset
            ):
                for temp_offset in range(0, temporal_size, 2):
                    slice_offsets.append(
                        np.arange(
                            b_offset + temp_offset * spatial_size,
                            b_offset + (temp_offset + 1) * spatial_size,
                        )
                    )
            slice_offsets = np.concatenate(slice_offsets, axis=-1).astype(np.int32)

            slice_offsets2 = []
            for temporal_size, spatial_size, b_offset in zip(
                grid_t, grid_hw_after_conv, batch_offset
            ):
                for temp_offset in range(
                    1 if temporal_size > 1 else 0, temporal_size, 2
                ):
                    slice_offsets2.append(
                        np.arange(
                            b_offset + temp_offset * spatial_size,
                            b_offset + (temp_offset + 1) * spatial_size,
                        )
                    )
            slice_offsets2 = np.concatenate(slice_offsets2, axis=-1).astype(np.int32)

            x_timestep_1 = x[mx.array(slice_offsets), :]
            x_timestep_2 = x[mx.array(slice_offsets2), :]
            x = mx.concatenate([x_timestep_1, x_timestep_2], axis=-1)
            return x

        def fwd_temporal(x):
            x = self.temporal_linear(x)
            return x

        def fwd_mlp(x):
            x = self.mlp(x)
            x = self.after_norm(x)
            return x

        x = fwd_spatial(x)
        if self.use_temporal_conv:
            x = fwd_placeholder(x, grid_thw)
            x = fwd_temporal(x)
        x = fwd_mlp(x)
        return x


class Model(nn.Module):
    """ERNIE 4.5 VL MoE model."""

    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.resampler_model = VariableResolutionResamplerModel(
            config.pixel_hidden_size,
            config.hidden_size,
            config.spatial_conv_size,
            config.temporal_conv_size,
            config=config,
        )
        self.language_model = LanguageModel(config.text_config, config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        image_grid_thw = kwargs.get("image_grid_thw", None)
        video_grid_thw = kwargs.get("video_grid_thw", None)
        grid_thw = image_grid_thw if image_grid_thw is not None else video_grid_thw

        if pixel_values is None:
            return InputEmbeddingsFeatures(
                inputs_embeds=self.language_model.model.embed_tokens(input_ids)
            )

        dtype = self.vision_tower.patch_embed.proj.weight.dtype
        pixel_values = pixel_values.astype(dtype)

        inputs_embeds = self.language_model.model.embed_tokens(input_ids)
        hidden_states = self.vision_tower(
            pixel_values, grid_thw, output_hidden_states=False
        )
        image_features = self.resampler_model(hidden_states, image_grid_thw)
        final_inputs_embeds = self._merge_input_ids_with_image_features(
            image_features,
            inputs_embeds,
            input_ids,
        )
        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    def _merge_input_ids_with_image_features(
        self,
        image_features: mx.array,
        inputs_embeds: mx.array,
        input_ids: mx.array,
    ) -> mx.array:
        image_token_id = self.config.image_token_id
        video_token_id = self.config.video_token_id

        image_positions = input_ids == image_token_id
        if mx.sum(image_positions) == 0:
            image_positions = input_ids == video_token_id

        if mx.sum(image_positions) == 0:
            return inputs_embeds

        batch_size, seq_len = input_ids.shape
        batch_outputs = []
        feature_start_idx = 0

        for batch_idx in range(batch_size):
            image_mask = image_positions[batch_idx]
            num_positions = int(mx.sum(image_mask).item())

            if num_positions > 0:
                batch_features = image_features[
                    feature_start_idx : feature_start_idx + num_positions
                ]

                if batch_features.shape[0] != num_positions:
                    raise ValueError(
                        f"Number of image token positions ({num_positions}) does not match "
                        f"number of image features ({batch_features.shape[0]}) for batch {batch_idx}"
                    )

                cumsum = mx.cumsum(image_mask.astype(mx.int32))
                feature_indices = mx.where(
                    image_mask, cumsum - 1, mx.zeros_like(cumsum)
                )
                gathered_features = batch_features[feature_indices]

                image_mask_expanded = mx.expand_dims(image_mask, axis=-1)
                batch_output = mx.where(
                    image_mask_expanded, gathered_features, inputs_embeds[batch_idx]
                )

                feature_start_idx += num_positions
            else:
                batch_output = inputs_embeds[batch_idx]

            batch_outputs.append(batch_output)

        return mx.stack(batch_outputs, axis=0)

    @property
    def layers(self):
        return self.language_model.model.layers

    def _build_token_type_ids(
        self, input_ids: mx.array, pixel_values: Optional[mx.array] = None
    ) -> Optional[mx.array]:
        if pixel_values is None:
            return None

        image_token_id = self.config.image_token_id
        video_token_id = self.config.video_token_id

        is_image = input_ids == image_token_id
        is_video = input_ids == video_token_id
        is_vision = is_image | is_video

        if mx.sum(is_vision) == 0:
            return None

        token_type_ids = mx.where(
            is_vision, mx.ones_like(input_ids), mx.zeros_like(input_ids)
        )
        return token_type_ids

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):
        token_type_ids = self._build_token_type_ids(input_ids, pixel_values)

        inputs_embeds_features = self.get_input_embeddings(
            input_ids, pixel_values, **kwargs
        )

        logits = self.language_model(
            input_ids,
            inputs_embeds_features.inputs_embeds,
            mask=mask,
            cache=cache,
            pixel_values=pixel_values,
            token_type_ids=token_type_ids,
            **kwargs,
        )

        return logits

    def sanitize(self, weights):
        import re

        def transform_key(key):
            if "vision_tower" not in key and "vision_model" in key:
                key = key.replace("vision_model", "vision_tower")

            if "language_model" not in key:
                if (
                    "model.layers" in key
                    or "model.embed_tokens" in key
                    or "model.norm" in key
                ):
                    key = key.replace("model.", "language_model.model.")
                elif "lm_head" in key:
                    key = key.replace("lm_head", "language_model.lm_head")

            if "model.resampler_model" in key:
                key = key.replace("model.resampler_model", "resampler_model")

            key = re.sub(
                r"(spatial_linear|temporal_linear)\.(\d+)", r"\1.layers.\2", key
            )

            return key

        weights = {transform_key(k): v for k, v in weights.items()}
        weights = self.vision_tower.sanitize(weights)
        weights = self.language_model.sanitize(weights)

        return weights



"""
MLX-based KimiVL Processor.

This module provides an MLX-native processor for KimiVL models that:
1. Uses a pre-converted fast tokenizer (no tiktoken dependency)
2. Provides an MLX-based image processor (no torch/torchvision dependency)
3. Patches missing functions for transformers 5.0 compatibility
"""

import json
import math
import warnings
from pathlib import Path
from typing import List, Optional, Tuple, Union

import mlx.core as mx
import transformers.processing_utils as processing_utils
from PIL import Image
from transformers import AutoTokenizer
from transformers.feature_extraction_utils import BatchFeature
from transformers.image_processing_utils import BaseImageProcessor
from transformers.image_utils import ImageInput, make_list_of_images, valid_images
from transformers.processing_utils import ProcessorMixin
from transformers.tokenization_utils_base import PreTokenizedInput, TextInput
from transformers.utils import TensorType

from .config import ModelConfig


def _validate_images_text_input_order(images, text):
    """
    Validate and potentially swap the order of images and text arguments.

    This function checks if the arguments are in the correct order (images first, text second)
    for backward compatibility. If text is passed as the first argument and images as the second,
    it swaps them and issues a deprecation warning.

    Args:
        images: The images argument (should be image-like objects or None)
        text: The text argument (should be strings or None)

    Returns:
        Tuple of (images, text) in the correct order
    """
    # Check if arguments are swapped (text passed as images, images passed as text)
    if images is not None and text is not None:
        # If 'images' looks like text and 'text' looks like images, swap them
        images_is_text = isinstance(images, str) or (
            isinstance(images, (list, tuple))
            and len(images) > 0
            and isinstance(images[0], str)
        )
        text_is_image = not isinstance(text, str) and not (
            isinstance(text, (list, tuple))
            and len(text) > 0
            and isinstance(text[0], str)
        )

        if images_is_text and text_is_image:
            warnings.warn(
                "You passed text as the first argument and images as the second. "
                "This is deprecated and will be removed in a future version. "
                "Please pass images first and text second.",
                FutureWarning,
            )
            return text, images

    return images, text


# Add the function to transformers.processing_utils if it doesn't exist
if not hasattr(processing_utils, "_validate_images_text_input_order"):
    processing_utils._validate_images_text_input_order = (
        _validate_images_text_input_order
    )

# Also add Unpack if it doesn't exist (for older Python versions)
if not hasattr(processing_utils, "Unpack"):
    try:
        from typing import Unpack

        processing_utils.Unpack = Unpack
    except ImportError:
        from typing_extensions import Unpack

        processing_utils.Unpack = Unpack


def _ensure_gpt2_bytes_to_unicode():
    """
    Compatibility shim for tokenizer remote code expecting legacy GPT-2 helper.

    Newer Transformers versions removed bytes_to_unicode from
    transformers.models.gpt2.tokenization_gpt2, but remote tokenizers may still
    import it from there.
    """
    try:
        import transformers.models.gpt2.tokenization_gpt2 as gpt2_tokenization
    except Exception:
        return

    if hasattr(gpt2_tokenization, "bytes_to_unicode"):
        return

    def bytes_to_unicode():
        bs = (
            list(range(ord("!"), ord("~") + 1))
            + list(range(ord("¡"), ord("¬") + 1))
            + list(range(ord("®"), ord("ÿ") + 1))
        )
        cs = bs[:]
        n = 0
        for b in range(2**8):
            if b not in bs:
                bs.append(b)
                cs.append(2**8 + n)
                n += 1
        cs = [chr(n) for n in cs]
        return dict(zip(bs, cs))

    gpt2_tokenization.bytes_to_unicode = bytes_to_unicode


# CLIP-style normalization constants
OPENAI_DATASET_MEAN = (0.48145466, 0.4578275, 0.40821073)
OPENAI_DATASET_STD = (0.26862954, 0.26130258, 0.27577711)


class KimiVLImageProcessor(BaseImageProcessor):

    model_input_names = ["pixel_values", "image_grid_hws"]

    def __init__(
        self,
        patch_size: int = 14,
        pad_input: bool = False,
        image_mean: Tuple[float, float, float] = OPENAI_DATASET_MEAN,
        image_std: Tuple[float, float, float] = OPENAI_DATASET_STD,
        in_token_limit: int = 4096,
        merge_kernel_size: List[int] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.in_token_limit = in_token_limit
        self.patch_size = patch_size
        self.pad_input = pad_input
        self.image_mean = image_mean
        self.image_std = image_std
        self.merge_kernel_size = (
            merge_kernel_size if merge_kernel_size is not None else [2, 2]
        )

    def rescale(
        self, image: Image.Image, merge_kernel_size: List[int] = None
    ) -> Image.Image:
        """Rescale image to fit within token limits and pad/crop to patch boundaries."""
        if merge_kernel_size is None:
            merge_kernel_size = self.merge_kernel_size

        w, h = image.size
        patch_size = self.patch_size

        # Rescale if exceeds token limit
        if (w // patch_size) * (h // patch_size) > self.in_token_limit:
            scale = math.sqrt(
                self.in_token_limit / ((w // patch_size) * (h // patch_size))
            )
            new_w, new_h = int(w * scale), int(h * scale)
            image = image.resize((new_w, new_h), Image.Resampling.BICUBIC)

        if self.pad_input:
            new_w, new_h = image.size
            pad_size_h = merge_kernel_size[0] * patch_size
            pad_size_w = merge_kernel_size[1] * patch_size

            pad_h = (pad_size_h - new_h % pad_size_h) % pad_size_h
            pad_w = (pad_size_w - new_w % pad_size_w) % pad_size_w

            if pad_h > 0 or pad_w > 0:
                # Pad image (bottom and right padding)
                new_image = Image.new(
                    image.mode, (new_w + pad_w, new_h + pad_h), (0, 0, 0)
                )
                new_image.paste(image, (0, 0))
                image = new_image
        else:
            new_w, new_h = image.size
            # Ensure dimensions are divisible by merge_kernel_size * patch_size
            # so that the grid dimensions are divisible by merge_kernel_size
            crop_size_w = merge_kernel_size[1] * patch_size
            crop_size_h = merge_kernel_size[0] * patch_size
            new_w = new_w - new_w % crop_size_w
            new_h = new_h - new_h % crop_size_h
            # Center crop
            left = (image.size[0] - new_w) // 2
            top = (image.size[1] - new_h) // 2
            image = image.crop((left, top, left + new_w, top + new_h))

        w, h = image.size
        if w // patch_size >= 512 or h // patch_size >= 512:
            raise ValueError("Exceed pos emb")

        return image

    def to_mlx(self, image: Image.Image) -> mx.array:
        """Convert PIL image to MLX array in CHW format, normalized to [0, 1]."""
        image = image.convert("RGB")
        w, h = image.size
        # Convert PIL image to MLX array directly via bytes
        arr = mx.array(list(image.getdata()), dtype=mx.float32).reshape(h, w, 3) / 255.0
        # Convert from HWC to CHW format
        arr = arr.transpose(2, 0, 1)
        return arr

    def normalize(self, image: mx.array) -> mx.array:
        """Normalize image with CLIP-style mean and std."""
        mean = mx.array(self.image_mean, dtype=mx.float32).reshape(3, 1, 1)
        std = mx.array(self.image_std, dtype=mx.float32).reshape(3, 1, 1)
        return (image - mean) / std

    def patchify(self, image: mx.array) -> Tuple[mx.array, Tuple[int, int]]:
        """Convert image to patches."""
        patch_size = self.patch_size
        C, H, W = image.shape

        # Reshape to (C, H//p, p, W//p, p) then to (num_patches, C, p, p)
        patches = image.reshape(
            C, H // patch_size, patch_size, W // patch_size, patch_size
        )
        # Permute to (H//p, W//p, C, p, p)
        patches = patches.transpose(1, 3, 0, 2, 4)
        # Flatten to (num_patches, C, p, p)
        patches = patches.reshape(-1, C, patch_size, patch_size)

        grid_hw = (H // patch_size, W // patch_size)
        return patches, grid_hw

    def _preprocess(self, image: ImageInput) -> Tuple[mx.array, Tuple[int, int]]:
        """
        Preprocess image and patchify it.

        Args:
            image: Image to preprocess.

        Returns:
            patches: mx.array
            grid_hw: Tuple[int, int]
        """
        image = self.rescale(image, self.merge_kernel_size)
        image = self.to_mlx(image)
        image = self.normalize(image)
        patches, grid_hw = self.patchify(image)
        return patches, grid_hw

    def preprocess(
        self,
        images: ImageInput,
        return_tensors: Optional[Union[str, TensorType]] = None,
        **kwargs,
    ) -> BatchFeature:
        """Process images and return BatchFeature."""
        images = make_list_of_images(images)

        if not valid_images(images):
            raise ValueError(
                "Invalid image type. Must be of type PIL.Image.Image or mx.array."
            )

        pixel_values_list = []
        image_grid_hws = []

        for image in images:
            # Convert MLX arrays to PIL Images if needed
            if isinstance(image, mx.array):
                # Ensure we're working with the array values
                arr = image
                if arr.ndim == 3 and arr.shape[0] in [1, 3, 4]:
                    # CHW format, convert to HWC
                    arr = arr.transpose(1, 2, 0)
                # Convert to uint8 for PIL
                if arr.dtype in [mx.float32, mx.float16, mx.bfloat16]:
                    arr = (arr * 255).astype(mx.uint8)
                # Convert to PIL via list (MLX -> list -> PIL)
                h, w, _ = arr.shape
                flat_data = arr.reshape(-1).tolist()
                image = Image.frombytes("RGB", (w, h), bytes(flat_data))

            patches, image_grid_hw = self._preprocess(image)
            pixel_values_list.append(patches)
            image_grid_hws.append(image_grid_hw)

        pixel_values = mx.concatenate(pixel_values_list, axis=0)
        image_grid_hws = mx.array(image_grid_hws)

        # Return MLX arrays directly
        data = {
            "pixel_values": pixel_values,
            "image_grid_hws": image_grid_hws,
        }

        return BatchFeature(data=data, tensor_type=return_tensors)

    def __call__(
        self,
        images: ImageInput,
        return_tensors: Optional[Union[str, TensorType]] = None,
        **kwargs,
    ) -> BatchFeature:
        """Make the image processor callable."""
        return self.preprocess(images, return_tensors=return_tensors, **kwargs)


class KimiVLProcessor(ProcessorMixin):
    """
    MLX-based processor for KimiVL that doesn't require torch/torchvision.

    Constructs a KimiVL processor which wraps a KimiVL image processor and a tokenizer
    into a single processor.
    """

    attributes = ["image_processor", "tokenizer"]
    valid_kwargs = ["chat_template"]
    image_processor_class = "KimiVLImageProcessor"
    tokenizer_class = "AutoTokenizer"

    def __init__(
        self,
        image_processor=None,
        tokenizer=None,
        chat_template=None,
        **kwargs,
    ):
        self.image_token = "<|media_pad|>"
        if image_processor is None:
            image_processor = KimiVLImageProcessor()
        super().__init__(image_processor, tokenizer, chat_template=chat_template)

    def __call__(
        self,
        images: ImageInput = None,
        text: Union[
            TextInput, PreTokenizedInput, List[TextInput], List[PreTokenizedInput]
        ] = None,
        **kwargs,
    ) -> BatchFeature:
        """
        Main method to prepare for the model one or several sequences(s) and image(s).

        Args:
            images: The image or batch of images to be prepared.
            text: The sequence or batch of sequences to be encoded.
            return_tensors: If set, will return tensors of a particular framework.

        Returns:
            BatchFeature with input_ids, attention_mask, and pixel_values.
        """
        if images is None and text is None:
            raise ValueError("You have to specify at least one of `images` or `text`.")

        # Check if images and text inputs are reversed for BC
        images, text = _validate_images_text_input_order(images, text)

        # Extract return_tensors from kwargs (unused, we always return MLX arrays)
        kwargs.pop("return_tensors", None)

        # Process images
        if images is not None:
            image_inputs = self.image_processor(images)
            image_grid_hws = image_inputs["image_grid_hws"]
        else:
            image_inputs = {}
            image_grid_hws = None

        # Process text
        if isinstance(text, str):
            text = [text]
        elif text is not None and not isinstance(text, list):
            raise ValueError(
                "Invalid input text. Please provide a string, or a list of strings"
            )

        # Replace image tokens with the correct number of placeholder tokens
        if image_grid_hws is not None and text is not None:
            merge_length = (
                self.image_processor.merge_kernel_size[0]
                * self.image_processor.merge_kernel_size[1]
            )
            index = 0
            for i in range(len(text)):
                while self.image_token in text[i]:
                    # Use mx.prod for MLX arrays
                    grid_hw = image_grid_hws[index]
                    num_placeholders = int(mx.prod(grid_hw).item()) // merge_length
                    text[i] = text[i].replace(
                        self.image_token,
                        "<|placeholder|>" * num_placeholders,
                        1,
                    )
                    index += 1
                text[i] = text[i].replace("<|placeholder|>", self.image_token)

        # Tokenize text
        # Note: The TikToken tokenizer doesn't work properly with transformers' standard
        # __call__ method due to issues with the pad function. We use encode() directly.
        if text is not None:
            # Encode each text and build the result manually
            all_input_ids = []
            for t in text:
                ids = self.tokenizer.encode(t)
                all_input_ids.append(ids)

            # Pad sequences to the same length if needed
            max_len = max(len(ids) for ids in all_input_ids)
            pad_token_id = self.tokenizer.pad_token_id or 0

            padded_input_ids = []
            attention_masks = []
            for ids in all_input_ids:
                padding_length = max_len - len(ids)
                padded_ids = ids + [pad_token_id] * padding_length
                mask = [1] * len(ids) + [0] * padding_length
                padded_input_ids.append(padded_ids)
                attention_masks.append(mask)

            # Convert to MLX arrays
            text_inputs = {
                "input_ids": mx.array(padded_input_ids),
                "attention_mask": mx.array(attention_masks),
            }
        else:
            text_inputs = {}

        data = {**text_inputs, **image_inputs}
        image_token_id = self.tokenizer.convert_tokens_to_ids(self.image_token)
        if image_token_id is not None:
            data["image_token_id"] = int(image_token_id)
        return BatchFeature(data=data)

    def batch_decode(self, *args, **kwargs):
        """Forward to tokenizer's batch_decode."""
        return self.tokenizer.batch_decode(*args, **kwargs)

    def decode(self, *args, **kwargs):
        """Forward to tokenizer's decode."""
        return self.tokenizer.decode(*args, **kwargs)

    def apply_chat_template(
        self,
        conversation,
        chat_template=None,
        add_generation_prompt=False,
        tokenize=False,
        **kwargs,
    ):
        """Apply chat template to the conversation."""
        # Use provided template, processor's template, or tokenizer's template
        if chat_template is None:
            chat_template = self.chat_template
        if chat_template is None:
            chat_template = getattr(self.tokenizer, "chat_template", None)
        if chat_template is None:
            raise ValueError(
                "No chat template found. Please provide a chat_template argument "
                "or ensure the tokenizer has a chat_template attribute."
            )

        # Use jinja2 to render the template
        try:
            from jinja2 import Template
        except ImportError:
            raise ImportError("jinja2 is required for apply_chat_template")

        template = Template(chat_template)
        rendered = template.render(
            messages=conversation,
            add_generation_prompt=add_generation_prompt,
            **kwargs,
        )

        if tokenize:
            return self.tokenizer.encode(rendered)
        return rendered

    @property
    def model_input_names(self):
        """Get the model input names from tokenizer and image processor."""
        tokenizer_input_names = self.tokenizer.model_input_names
        image_processor_input_names = self.image_processor.model_input_names
        return list(dict.fromkeys(tokenizer_input_names + image_processor_input_names))

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, **kwargs):
        """Load the processor from a pretrained model path."""
        from huggingface_hub import hf_hub_download

        kwargs.pop("trust_remote_code", None)
        _ensure_gpt2_bytes_to_unicode()

        model_path = Path(pretrained_model_name_or_path)
        is_local = model_path.exists() and model_path.is_dir()
        tokenizer = AutoTokenizer.from_pretrained(
            str(model_path) if is_local else pretrained_model_name_or_path,
            trust_remote_code=True,
            local_files_only=is_local,
        )

        # Load image processor config and create our processor
        image_processor_config = {}
        try:
            if is_local:
                config_path = model_path / "config.json"
            else:
                config_path = Path(
                    hf_hub_download(pretrained_model_name_or_path, "config.json")
                )
            with open(config_path, "r", encoding="utf-8") as f:
                config_dict = json.load(f)
            config = ModelConfig.from_dict(config_dict)
            if hasattr(config, "vision_config"):
                vision_config = config.vision_config
                if hasattr(vision_config, "patch_size"):
                    image_processor_config["patch_size"] = vision_config.patch_size
                if hasattr(vision_config, "in_token_limit"):
                    image_processor_config["in_token_limit"] = (
                        vision_config.in_token_limit
                    )
                if hasattr(vision_config, "merge_kernel_size"):
                    image_processor_config["merge_kernel_size"] = (
                        vision_config.merge_kernel_size
                    )
        except Exception:
            pass

        image_processor = KimiVLImageProcessor(**image_processor_config)

        # Load chat template from jinja file if not already set on tokenizer
        chat_template = getattr(tokenizer, "chat_template", None)
        if chat_template is None:
            try:
                if is_local:
                    jinja_path = model_path / "chat_template.jinja"
                else:
                    jinja_path = Path(
                        hf_hub_download(
                            pretrained_model_name_or_path, "chat_template.jinja"
                        )
                    )
                if jinja_path.exists():
                    chat_template = jinja_path.read_text(encoding="utf-8")
                    # Set chat_template on tokenizer so apply_chat_template works
                    tokenizer.chat_template = chat_template
            except Exception:
                pass

        return cls(
            image_processor=image_processor,
            tokenizer=tokenizer,
            chat_template=chat_template,
        )


from transformers import AutoProcessor

_original_auto_processor_from_pretrained = AutoProcessor.from_pretrained


@classmethod
def _patched_auto_processor_from_pretrained(
    cls, pretrained_model_name_or_path, **kwargs
):
    """Patched from_pretrained that returns KimiVLProcessor for kimi_vl models."""
    from huggingface_hub import hf_hub_download

    model_path = Path(pretrained_model_name_or_path)
    is_local = model_path.exists() and model_path.is_dir()

    # Check if this is a kimi_vl model
    is_kimi_vl = False
    try:
        if is_local:
            config_path = model_path / "config.json"
        else:
            config_path = Path(
                hf_hub_download(pretrained_model_name_or_path, "config.json")
            )
        with open(config_path, "r", encoding="utf-8") as f:
            config = json.load(f)
        is_kimi_vl = config.get("model_type", "").lower() == "kimi_vl"
    except Exception:
        pass

    if is_kimi_vl:
        return KimiVLProcessor.from_pretrained(pretrained_model_name_or_path, **kwargs)

    return _original_auto_processor_from_pretrained.__func__(
        cls, pretrained_model_name_or_path, **kwargs
    )


AutoProcessor.from_pretrained = _patched_auto_processor_from_pretrained



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from ..base import InputEmbeddingsFeatures
from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


class LlavaMultiModalProjector(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.linear_1 = nn.Linear(
            config.vision_config.hidden_size, config.text_config.hidden_size, bias=True
        )
        self.gelu = nn.GELU()
        self.linear_2 = nn.Linear(
            config.text_config.hidden_size, config.text_config.hidden_size, bias=True
        )

    def __call__(self, x: mx.array) -> mx.array:
        x = self.linear_1(x)
        x = self.gelu(x)
        x = self.linear_2(x)
        return x


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        embed_std = 1 / mx.sqrt(config.text_config.hidden_size)
        self.image_newline = (
            mx.random.normal((config.text_config.hidden_size,)) * embed_std
        )

        self.multi_modal_projector = LlavaMultiModalProjector(config)
        self.vision_feature_layer = config.vision_feature_layer
        self.vision_feature_select_strategy = config.vision_feature_select_strategy

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        if pixel_values is None:
            return InputEmbeddingsFeatures(
                inputs_embeds=self.language_model.model.embed_tokens(input_ids)
            )

        # Get the input embeddings from the language model
        inputs_embeds = self.language_model.model.embed_tokens(input_ids)

        # Get the ouptut hidden states from the vision model
        *_, hidden_states = self.vision_tower(
            pixel_values[0].transpose(0, 2, 3, 1), output_hidden_states=True
        )

        # Select the hidden states from the desired layer
        selected_image_feature = hidden_states[self.vision_feature_layer]

        if self.vision_feature_select_strategy == "default":
            selected_image_feature = selected_image_feature[:, 1:]
        elif self.vision_feature_select_strategy == "full":
            selected_image_feature = selected_image_feature
        else:
            raise ValueError(
                "Unexpected feature selection strategy: "
                f"{self.vision_feature_select_strategy}"
            )

        # Pass image features through the multi-modal projector
        image_features = self.multi_modal_projector(selected_image_feature)

        # Add a newline token to the image features
        if self.image_newline is not None:
            self.image_newline = np.array(self.image_newline)[None, None, :]
            self.image_newline = np.broadcast_to(
                self.image_newline, image_features.shape
            )
            image_newline = mx.array(self.image_newline)
            image_features = mx.concatenate([image_features, image_newline], axis=0)

        # Insert special image tokens in the input_ids
        final_inputs_embeds = self._merge_input_ids_with_image_features(
            image_features, inputs_embeds, input_ids
        )
        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    def _merge_input_ids_with_image_features(
        self, image_features, inputs_embeds, input_ids
    ):
        image_token_index = self.config.image_token_index
        num_images, num_image_patches, embed_dim = image_features.shape

        image_positions = np.where(input_ids == image_token_index)[1].tolist()

        text_segments = []
        start_idx = 0

        for position in image_positions:
            text_segments.append(inputs_embeds[:, start_idx:position])
            start_idx = position + 1

        image_embeddings = mx.split(image_features, image_features.shape[0])
        final_embeddings = [v for p in zip(text_segments, image_embeddings) for v in p]
        final_embeddings += [inputs_embeds[:, start_idx:]]

        # Create a final embedding of shape
        # (1, num_image_patches*num_images + sequence_len, embed_dim)
        return mx.concatenate(final_embeddings, axis=1)

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        mask: mx.array,
        cache=None,
        **kwargs,
    ):

        input_embeddings_features = self.get_input_embeddings(input_ids, pixel_values)
        logits = self.language_model(
            input_ids,
            cache=cache,
            inputs_embeds=input_embeddings_features.inputs_embeds,
        )
        return logits



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from mlx_lm.models.rope_utils import SuScaledRoPE

from ..base import InputEmbeddingsFeatures, LanguageModelOutput, create_attention_mask
from ..cache import KVCache

# Import processor to register it with AutoProcessor
from . import processing_phi3_v  # noqa: F401
from .config import ModelConfig, TextConfig
from .vision import VisionModel


class Attention(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()

        dim = config.hidden_size
        self.n_heads = n_heads = config.num_attention_heads
        self.n_kv_heads = n_kv_heads = config.num_key_value_heads
        self.num_hidden_layers = config.num_hidden_layers

        self.head_dim = head_dim = config.hidden_size // n_heads
        self.scale = head_dim**-0.5

        op_size = n_heads * head_dim + 2 * (n_kv_heads * head_dim)
        self.qkv_proj = nn.Linear(dim, op_size, bias=False)
        self.o_proj = nn.Linear(n_heads * head_dim, dim, bias=False)

        rope_dim = int(head_dim * config.partial_rotary_factor)

        # Check for Su-scaled RoPE by type or presence of short/long factors
        rope_type = config.rope_scaling.get("type") if config.rope_scaling else None
        has_su_factors = (
            config.rope_scaling
            and "short_factor" in config.rope_scaling
            and "long_factor" in config.rope_scaling
        )

        if rope_type == "su" or has_su_factors:
            self.rope = SuScaledRoPE(
                rope_dim,
                base=config.rope_theta,
                max_position_embeddings=config.max_position_embeddings,
                original_max_position_embeddings=config.original_max_position_embeddings,
                short_factor=config.rope_scaling["short_factor"],
                long_factor=config.rope_scaling["long_factor"],
            )
        else:
            rope_scale = 1.0
            if config.rope_scaling and rope_type == "linear":
                rope_scale = 1 / config.rope_scaling["factor"]
            self.rope = nn.RoPE(
                rope_dim,
                traditional=config.rope_traditional,
                base=config.rope_theta,
                scale=rope_scale,
            )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        B, L, _ = x.shape

        qkv = self.qkv_proj(x)
        query_pos = self.n_heads * self.head_dim
        queries, keys, values = mx.split(
            qkv, [query_pos, query_pos + self.n_kv_heads * self.head_dim], axis=-1
        )

        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        if cache is not None:
            queries = self.rope(queries, offset=cache.offset)
            keys = self.rope(keys, offset=cache.offset)
            keys, values = cache.update_and_fetch(keys, values)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        output = mx.fast.scaled_dot_product_attention(
            queries, keys, values, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.o_proj(output)


class MLP(nn.Module):
    def __init__(self, dim, hidden_dim):
        super().__init__()
        self.gate_up_proj = nn.Linear(dim, 2 * hidden_dim, bias=False)
        self.down_proj = nn.Linear(hidden_dim, dim, bias=False)

    def __call__(self, x) -> mx.array:
        x = self.gate_up_proj(x)
        gate, x = mx.split(x, 2, axis=-1)
        return self.down_proj(nn.silu(gate) * x)


class TransformerBlock(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.num_attention_heads = config.num_attention_heads
        self.hidden_size = config.hidden_size
        self.self_attn = Attention(config)
        self.mlp = MLP(config.hidden_size, config.intermediate_size)
        self.input_layernorm = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.post_attention_layernorm = nn.RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.config = config

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        r = self.self_attn(self.input_layernorm(x), mask, cache)
        h = x + r
        r = self.mlp(self.post_attention_layernorm(h))
        out = h + r
        return out


class Phi3V(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vocab_size = config.vocab_size
        self.num_hidden_layers = config.num_hidden_layers
        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size)
        self.vision_embed_tokens = VisionModel(config)
        self.layers = [
            TransformerBlock(config=config) for _ in range(config.num_hidden_layers)
        ]
        self.norm = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
    ):
        if inputs_embeds is None:
            h = self.embed_tokens(inputs)
        else:
            h = inputs_embeds

        if cache is None:
            cache = [None] * len(self.layers)

        if mask is None:
            mask = create_attention_mask(h, cache[0])

        for layer, c in zip(self.layers, cache):
            h = layer(h, mask, c)

        return self.norm(h)


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.model_type = config.model_type
        self.model = Phi3V(config)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.config = config

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        pixel_values=None,
        mask=None,
        cache=None,
        **kwargs,
    ):
        if inputs_embeds is None:
            input_embeddings_features = self.get_input_embeddings(
                inputs, pixel_values, **kwargs
            )
            inputs_embeds = input_embeddings_features.inputs_embeds

        out = self.model(inputs, inputs_embeds, mask=mask, cache=cache)
        logits = self.lm_head(out)

        return LanguageModelOutput(logits=logits)

    def get_input_embeddings(
        self,
        inputs: mx.array,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        image_sizes = kwargs.get("image_sizes", None) if kwargs else None

        # Get text embeddings
        inputs_embeds = self.model.embed_tokens(inputs)

        # Find positions where inputs < 0 (image token positions)
        inputs_list = inputs.tolist()
        p = np.argwhere(np.array(inputs_list) < 0).tolist()

        if pixel_values is not None:
            if pixel_values.dtype != inputs_embeds.dtype:
                pixel_values = pixel_values.astype(inputs_embeds.dtype)

            inputs_embeds = self.model.vision_embed_tokens(
                pixel_values, inputs_embeds, image_sizes, p
            )

        return InputEmbeddingsFeatures(inputs_embeds=inputs_embeds)

    @property
    def layers(self):
        return self.model.layers

    @property
    def head_dim(self):
        return self.config.hidden_size // self.config.num_attention_heads

    @property
    def n_kv_heads(self):
        return self.config.num_key_value_heads

    @property
    def language_model(self):
        return self

    @property
    def vision_model(self):
        return self.model.vision_embed_tokens



"""Processor for Jina VLM in MLX-VLM."""

from typing import Dict, List, Literal, Optional, Union

import mlx.core as mx
import numpy as np
from PIL import Image
from transformers.processing_utils import ProcessorMixin

from .image_processor import ImageProcessor


class JinaVLMProcessor(ProcessorMixin):
    """Processor for Jina VLM that combines tokenizer and image processor."""

    tokenizer_class = ("Qwen2Tokenizer", "Qwen2TokenizerFast")
    attributes = ["tokenizer"]

    def __init__(
        self,
        tokenizer,
        image_token: str = "<|image|>",
        chat_template: Optional[str] = None,
        **kwargs,
    ):
        self.tokenizer = tokenizer
        self.image_token = image_token
        self._image_proc = ImageProcessor()  # Internal, not exposed as image_processor

        # Get image token ID
        self.image_token_id = self.tokenizer.convert_tokens_to_ids(image_token)

        super().__init__(tokenizer, **kwargs)

        # Set chat template AFTER super().__init__ - always set the default if not already set
        default_chat_template = (
            "{% for message in messages %}"
            "{% if message['role'] == 'system' %}{{ '<|system|>\n' + message['content'] + '\n' }}"
            "{% elif message['role'] == 'user' %}{{ '<|user|>\n' + message['content'] + '\n' }}"
            "{% elif message['role'] == 'assistant' %}{{ '<|assistant|>\n' + message['content'] + '\n' }}"
            "{% endif %}{% endfor %}"
            "{% if add_generation_prompt %}{{ '<|assistant|>\n' }}{% endif %}"
        )
        if chat_template is not None:
            self.tokenizer.chat_template = chat_template
        elif not self.tokenizer.chat_template:
            self.tokenizer.chat_template = default_chat_template

    @property
    def chat_template(self):
        return self.tokenizer.chat_template

    @chat_template.setter
    def chat_template(self, value):
        self.tokenizer.chat_template = value

    @property
    def pad_token(self):
        return self.tokenizer.pad_token

    @property
    def pad_token_id(self):
        return self.tokenizer.pad_token_id

    @property
    def eos_token(self):
        return self.tokenizer.eos_token

    @property
    def eos_token_id(self):
        return self.tokenizer.eos_token_id

    @property
    def bos_token(self):
        return self.tokenizer.bos_token

    @property
    def bos_token_id(self):
        return self.tokenizer.bos_token_id

    def encode(self, text: str, add_special_tokens: bool = False) -> List[int]:
        return self.tokenizer.encode(text, add_special_tokens=add_special_tokens)

    def decode(self, token_ids: List[int], **kwargs) -> str:
        return self.tokenizer.decode(token_ids, **kwargs)

    def batch_decode(self, token_ids, **kwargs) -> List[str]:
        return self.tokenizer.batch_decode(token_ids, **kwargs)

    def process_one(
        self,
        prompt: str,
        images: Optional[List[Image.Image]] = None,
        inference_mode: bool = True,
    ) -> Dict:
        """Process a single prompt with images."""
        if images is None:
            images = []

        # Process images
        if images:
            image_outputs = self._image_proc.preprocess(images)
            pixel_values_list = image_outputs["pixel_values"]
            image_tokens = image_outputs["image_tokens"]
            image_input_idx_list = image_outputs["image_input_idx"]
            image_masks_list = image_outputs["image_masks"]
        else:
            pixel_values_list = None
            image_tokens = []
            image_input_idx_list = None
            image_masks_list = None

        # Split prompt by image token
        text_splits = prompt.split(self.image_token)

        # Build input_ids with image tokens interleaved
        input_ids = []
        current_image_idx = 0
        updated_image_input_idx = []

        for i, text_part in enumerate(text_splits):
            # Encode text part
            if text_part:
                text_tokens = self.encode(text_part, add_special_tokens=False)
                input_ids.extend(text_tokens)

            # Add image tokens if not the last split and we have images
            if i < len(text_splits) - 1 and current_image_idx < len(image_tokens):
                # Get image tokens for this image
                img_tokens = image_tokens[current_image_idx]
                # Offset image_input_idx by current position
                if image_input_idx_list is not None and current_image_idx < len(
                    image_input_idx_list
                ):
                    offset_idx = image_input_idx_list[current_image_idx] + len(
                        input_ids
                    )
                    updated_image_input_idx.append(offset_idx)
                input_ids.extend(img_tokens.tolist())
                current_image_idx += 1

        input_ids = mx.array(input_ids)

        result = {
            "input_ids": input_ids[None, :],  # Add batch dimension
            "attention_mask": mx.ones_like(input_ids)[None, :],
        }

        if pixel_values_list is not None and len(pixel_values_list) > 0:
            # Stack pixel values: (n_crops, n_patches, patch_dim)
            result["pixel_values"] = mx.array(np.stack(pixel_values_list))
            # Stack image_input_idx: (n_images, tokens_per_image)
            result["image_input_idx"] = mx.array(np.stack(updated_image_input_idx))
            # Stack image_masks: (n_crops, n_patches)
            result["image_masks"] = mx.array(np.stack(image_masks_list))

        return result

    def __call__(
        self,
        text: Optional[Union[str, List[str]]] = None,
        images: Optional[Union[Image.Image, List[Image.Image]]] = None,
        inference_mode: bool = True,
        return_tensors: Literal["np", "mx", "pt"] = "mx",
        **kwargs,
    ) -> Dict:
        """Process text and images for Jina VLM.

        When called with just text (like a tokenizer), returns tokenizer output.
        When called with text and images, returns full processed inputs.

        Args:
            text: Input text or list of texts
            images: Input image or list of images
            inference_mode: Whether in inference mode
            return_tensors: Type of tensors to return

        Returns:
            Dictionary containing processed inputs
        """
        # If called with just text (like a tokenizer), delegate to tokenizer
        if text is not None and images is None:
            return self.tokenizer(text, **kwargs)

        if text is None:
            raise ValueError("Text must be provided")

        # Normalize inputs
        if isinstance(text, str):
            texts = [text]
        else:
            texts = text

        if images is None:
            images_list = [None] * len(texts)
        elif isinstance(images, Image.Image):
            images_list = [[images]]
        elif isinstance(images, list) and len(images) > 0:
            if isinstance(images[0], Image.Image):
                # Single list of images for single prompt
                images_list = [images]
            else:
                images_list = images
        else:
            images_list = [None] * len(texts)

        # Process each text-image pair
        batch_results = []
        for prompt, imgs in zip(texts, images_list):
            result = self.process_one(prompt, imgs, inference_mode)
            batch_results.append(result)

        # Collate results
        if len(batch_results) == 1:
            return batch_results[0]
        else:
            return self._collate_batch(batch_results)

    def _collate_batch(self, batch_results: List[Dict]) -> Dict:
        """Collate multiple results into a batch."""
        # Get max sequence length
        max_len = max(r["input_ids"].shape[1] for r in batch_results)

        padded_input_ids = []
        padded_attention_mask = []

        for r in batch_results:
            seq_len = r["input_ids"].shape[1]
            pad_len = max_len - seq_len

            if pad_len > 0:
                input_ids = mx.concatenate(
                    [mx.full((1, pad_len), self.pad_token_id), r["input_ids"]], axis=1
                )
                attention_mask = mx.concatenate(
                    [mx.zeros((1, pad_len)), r["attention_mask"]], axis=1
                )
            else:
                input_ids = r["input_ids"]
                attention_mask = r["attention_mask"]

            padded_input_ids.append(input_ids)
            padded_attention_mask.append(attention_mask)

        result = {
            "input_ids": mx.concatenate(padded_input_ids, axis=0),
            "attention_mask": mx.concatenate(padded_attention_mask, axis=0),
        }

        # Combine pixel values if present
        all_pixel_values = []
        all_image_input_idx = []
        all_image_masks = []

        for r in batch_results:
            if "pixel_values" in r:
                all_pixel_values.append(r["pixel_values"])
                all_image_input_idx.append(r["image_input_idx"])
                all_image_masks.append(r["image_masks"])

        if all_pixel_values:
            result["pixel_values"] = mx.concatenate(all_pixel_values, axis=0)
            result["image_input_idx"] = mx.concatenate(all_image_input_idx, axis=0)
            result["image_masks"] = mx.concatenate(all_image_masks, axis=0)

        return result
